# -*- coding: utf-8 -*-
"""Testing the functions in typhon.arts.
"""
import shutil

import pytest

from typhon import arts


class TestARTS:
    """Testing the ARTS utility functions."""
    @pytest.mark.skipif(not shutil.which('arts'), reason='arts not in PATH')
    def test_run_arts(self):
        """Test ARTS system call.

        Note: This test is only run, if ARTS is found in PATH.
        """
        arts_out = arts.run_arts(help=True)

        assert arts_out.retcode == 0



"""Utility functions related to plotting."""
import csv
import os
import re
from functools import lru_cache
from warnings import warn

import matplotlib.pyplot as plt
from matplotlib import colors
from matplotlib.colors import LinearSegmentedColormap
import numpy as np

from typhon.utils import deprecated

__all__ = [
    'cmap2rgba',
    'colors2cmap',
    'cmap2txt',
    'cmap2cpt',
    'cmap2act',
    'cmap2c3g',
    'cmap2ggr',
    'cmap_from_act',
    'cmap_from_txt',
    'get_material_design',
]


def cmap2rgba(cmap=None, N=None, interpolate=True):
    """Convert a colormap into a list of RGBA values.

    Parameters:
        cmap (str): Name of a registered colormap.
        N (int): Number of RGBA-values to return.
            If ``None`` use the number of colors defined in the colormap.
        interpolate (bool): Toggle the interpolation of values in the
            colormap.  If ``False``, only values from the colormap are
            used. This may lead to the re-use of a color, if the colormap
            provides less colors than requested. If ``True``, a lookup table
            is used to interpolate colors (default is ``True``).

    Returns:
        ndarray: RGBA-values.

    Examples:
        >>> cmap2rgba('viridis', 5)
        array([[ 0.267004,  0.004874,  0.329415,  1.      ],
            [ 0.229739,  0.322361,  0.545706,  1.      ],
            [ 0.127568,  0.566949,  0.550556,  1.      ],
            [ 0.369214,  0.788888,  0.382914,  1.      ],
            [ 0.993248,  0.906157,  0.143936,  1.      ]])
    """
    cmap = plt.get_cmap(cmap)

    if N is None:
        N = cmap.N

    nlut = N if interpolate else None

    if interpolate and isinstance(cmap, colors.ListedColormap):
        # `ListedColormap` does not support lookup table interpolation.
        cmap = colors.LinearSegmentedColormap.from_list('', cmap.colors)
        return cmap(np.linspace(0, 1, N))

    return plt.get_cmap(cmap.name, lut=nlut)(np.linspace(0, 1, N))


def _to_hex(c):
    """Convert arbitray color specification to hex string."""
    ctype = type(c)

    # Convert rgb to hex.
    if ctype is tuple or ctype is np.ndarray or ctype is list:
        return colors.rgb2hex(c)

    if ctype is str:
        # If color is already hex, simply return it.
        regex = re.compile('^#[A-Fa-f0-9]{6}$')
        if regex.match(c):
            return c

        # Convert named color to hex.
        return colors.cnames[c]

    raise Exception("Can't handle color of type: {}".format(ctype))


def colors2cmap(*args, name=None):
    """Create a colormap from a list of given colors.

    Parameters:
        *args: Arbitrary number of colors (Named color, HEX or RGB).
        name (str): Name with which the colormap is registered.

    Returns:
        LinearSegmentedColormap.

    Examples:
        >>> colors2cmap('darkorange', 'white', 'darkgreen', name='test')
    """
    if len(args) < 2:
        raise Exception("Give at least two colors.")

    cmap_data = [_to_hex(c) for c in args]
    cmap = colors.LinearSegmentedColormap.from_list(name, cmap_data)
    plt.register_cmap(name, cmap)

    return cmap


def cmap2txt(cmap, filename=None, N=None, comments='%'):
    """Export colormap to txt file.

    Parameters:
        cmap (str): Colormap name.
        filename (str): Optional filename.
            Default: cmap + '.txt'
        N (int): Number of colors.
        comments (str): Character to start comments with.

    """
    colors = cmap2rgba(cmap, N)
    header = 'Colormap "{}"'.format(cmap)

    if filename is None:
        filename = cmap + '.txt'

    np.savetxt(filename, colors[:, :3], fmt='%.4f', header=header,
               comments=comments)


def cmap2cpt(cmap, filename=None, N=None):
    """Export colormap to cpt file.

    Parameters:
        cmap (str): Colormap name.
        filename (str): Optional filename.
            Default: cmap + '.cpt'
        N (int): Number of colors.

    """
    colors = cmap2rgba(cmap, N)
    header = ('# GMT palette "{}"\n'
              '# COLOR_MODEL = RGB\n'.format(cmap))

    left = '{:>3d} {:>3d} {:>3d} {:>3d}  '.format
    right = '{:>3d} {:>3d} {:>3d} {:>3d}\n'.format

    if filename is None:
        filename = cmap + '.cpt'

    with open(filename, 'w') as f:
        f.write(header)

        # For each level specify a ...
        for n in range(len(colors)):
            rgb = [int(c * 255) for c in colors[n, :3]]
            # ... start color ...
            f.write(left(n, *rgb))
            # ... and end color.
            f.write(right(n + 1, *rgb))


def cmap2act(cmap, filename=None, N=None):
    """Export colormap to Adobe Color Table file.

    Parameters:
        cmap (str): Colormap name.
        filename (str): Optional filename.
            Default: cmap + '.cpt'
        N (int): Number of colors.

    """
    if filename is None:
        filename = cmap + '.act'

    # If the number of color levels to export is not set...
    if N is None:
        # ... use the number of colors defined in the colormap.
        N = plt.get_cmap(cmap).N

    if N > 256:
        N = 256
        warn('Maximum number of colors is 256.')

    colors = cmap2rgba(cmap, N)[:, :3]

    rgb = np.zeros(256 * 3 + 2)
    rgb[:colors.size] = (colors.flatten() * 255).astype(np.uint8)
    rgb[768:770] = np.uint8(N // 2**8), np.uint8(N % 2**8)

    rgb.astype(np.uint8).tofile(filename)


def cmap2c3g(cmap, filename=None, N=None):
    """Export colormap ass CSS3 gradient.

    Parameters:
        cmap (str): Colormap name.
        filename (str): Optional filename.
            Default: cmap + '.cpt'
        N (int): Number of colors.

    """
    if filename is None:
        filename = cmap + '.c3g'

    colors = cmap2rgba(cmap, N)

    header = (
        '/*'
        '   CSS3 Gradient "{}"\n'
        '*/\n\n'
        'linear-gradient(\n'
        '  0deg,\n'
        ).format(cmap)

    color_spec = '  rgb({:>3d},{:>3d},{:>3d}) {:>8.3%}'.format

    with open(filename, 'w') as f:
        f.write(header)

        ncolors = len(colors)
        for n in range(ncolors):
            r, g, b = [int(c * 255) for c in colors[n, :3]]
            f.write(color_spec(r, g, b, n / (ncolors - 1)))
            if n < ncolors - 1:
                f.write(',\n')

        f.write('\n  );')


def cmap2ggr(cmap, filename=None, N=None):
    """Export colormap as GIMP gradient.

    Parameters:
        cmap (str): Colormap name.
        filename (str): Optional filename.
            Default: cmap + '.cpt'
        N (int): Number of colors.

    """
    if filename is None:
        filename = cmap + '.ggr'

    colors = cmap2rgba(cmap, N)
    header = ('GIMP Gradient\n'
              'Name: {}\n'
              '{}\n').format(cmap, len(colors) - 1)

    line = ('{:.6f} {:.6f} {:.6f} '  # start, middle, stop
            '{:.6f} {:.6f} {:.6f} {:.6f} '  # RGBA
            '{:.6f} {:.6f} {:.6f} {:.6f} '  # RGBA next level
            '0 0\n').format

    def idx(x):
        return x / (len(colors) - 1)

    with open(filename, 'w') as f:
        f.write(header)

        for n in range(len(colors) - 1):
            rgb = colors[n, :]
            rgb_next = colors[n + 1, :]
            f.write(line(idx(n), idx(n + 0.5), idx(n + 1), *rgb, *rgb_next))


def cmap_from_act(file, name=None):
    """Import colormap from Adobe Color Table file.

    Parameters:
        file (str): Path to act file.
        name (str): Colormap name. Defaults to filename without extension.

    Returns:
        LinearSegmentedColormap.
    """
    # Extract colormap name from filename.
    if name is None:
        name = os.path.splitext(os.path.basename(file))[0]

    # Read binary file and determine number of colors
    rgb = np.fromfile(file, dtype=np.uint8)
    if rgb.shape[0] >= 770:
        ncolors = rgb[768] * 2**8 + rgb[769]
    else:
        ncolors = 256

    colors = rgb[:ncolors*3].reshape(ncolors, 3) / 255

    # Create and register colormap...
    cmap = LinearSegmentedColormap.from_list(name, colors, N=ncolors)
    plt.register_cmap(cmap=cmap)  # Register colormap.

    # ... and the reversed colormap.
    cmap_r = LinearSegmentedColormap.from_list(
            name + '_r', np.flipud(colors), N=ncolors)
    plt.register_cmap(cmap=cmap_r)

    return cmap


def cmap_from_txt(file, name=None, N=-1, comments='%'):
    """Import colormap from txt file.

    Reads colormap data (RGB/RGBA) from an ASCII file.
    Values have to be given in [0, 1] range.

    Parameters:
        file (str): Path to txt file.
        name (str): Colormap name. Defaults to filename without extension.
        N (int): Number of colors.
            ``-1`` means all colors (i.e., the complete file).
        comments (str): Character to start comments with.

    Returns:
        LinearSegmentedColormap.
    """
    # Extract colormap name from filename.
    if name is None:
        name = os.path.splitext(os.path.basename(file))[0]

    # Read binary file and determine number of colors
    rgb = np.genfromtxt(file, comments=comments)
    if N == -1:
        N = np.shape(rgb)[0]

    if np.min(rgb) < 0 or np.max(rgb) > 1:
        raise Exception('RGB value out of range: [0, 1].')

    # Create and register colormap...
    cmap = LinearSegmentedColormap.from_list(name, rgb, N=N)
    plt.register_cmap(cmap=cmap)

    # ... and the reversed colormap.
    cmap_r = LinearSegmentedColormap.from_list(
            name + '_r', np.flipud(rgb), N=N)
    plt.register_cmap(cmap=cmap_r)

    return cmap


@lru_cache(16)
def get_material_design(name, shade=None):
    """Return material design colors.

    Parameters:
        name (str): Color name (e.g. 'red').
        shade (str): Color shade (e.g. '500').
            If ``None`` all defined shades are returned.

    Returns:
        str or list[str]: Hex RGB value or list of hex RGB values.

    References:
        https://material.io/design/color/the-color-system.html

    Raises:
        ValueError: If the specified ``name`` or ``shade`` is not defined.

    Examples:
        >>> get_material_design('red', shade='500')
        '#F44336'

        >>> get_material_design('red')
        ['#FFEBEE', '#FFCDD2', '#EF9A9A', '#E57373', '#EF5350', '#F44336',
        '#E53935', '#D32F2F', '#C62828', '#B71C1C', '#FF8A80', '#FF5252',
        '#FF1744', '#D50000']

    """
    material_source = os.path.join(os.path.dirname(__file__), 'material.csv')
    with open(material_source, newline='') as csvfile:
        reader = csv.DictReader(csvfile, delimiter=',')

        available_colors = []
        for row in reader:
            available_colors.append(row['name'])
            if available_colors[-1] == name:
                available_shades = [k for k, v in row.items()
                                    if v and k != 'name']

                if shade is None:
                    return [c for c in
                            list(row.values())[1:len(available_shades) + 1]]

                if shade in available_shades:
                    return row[shade]
                else:
                    raise ValueError(
                        f'Shade "{shade}" not defined for color "{name}". '
                        f'Available shades are:\n{available_shades}'
                    )

        raise ValueError(
            f'Color "{name}" not defined. '
            f'Available colors are:\n{available_colors}.'
        )



# -*- coding: utf-8 -*-
"""Testing the functions in typhon.geographical.
"""
import numpy as np

from typhon import geographical


class TestGeographical:
    """Testing the geographical functions."""
    def test_area_weighted_mean(self):
        """Test calculation of area weighted mean."""
        lon = np.arange(0, 360, 10)
        lat = np.arange(-90, 90, 10)
        f = np.ones(lon.shape) + np.cos(np.deg2rad(lat))[:, np.newaxis]
        mean = geographical.area_weighted_mean(lon, lat, f)

        assert np.allclose(mean, 1.7852763105888174)


class TestGeoIndex:
    """Testing the GeoIndex functions."""

    def test_query(self):
        """Test the function collocate with dicts, xarray and GroupedArrays.
        Checks on temporal, spatial and temporal-spatial conditions.

        """
        # Update this test
        return

        # Test with dictionaries:
        lat1 = 30. * np.sin(np.linspace(-3.14, 3.14, 24)) + 20
        lon1 = np.linspace(0, 90, 24)
        lat2 = 30. * np.sin(np.linspace(-3.14, 3.14, 24) + 1.) + 20
        lon2 = np.linspace(0, 90, 24)

        index = geographical.GeoIndex(lat1, lon1)
        pairs, distances = index.query(lat2, lon2, r="500 km")

        check_pairs = [
            [4, 15, 15, 16],
            [4, 15, 16, 15]
        ]

        assert pairs.tolist() == check_pairs



# -*- coding: utf-8 -*-

"""Maths-related modules.
"""
from typhon.math import stats  # noqa
from typhon.math import array  # noqa
from typhon.math.common import *  # noqa

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
import copy
import numbers

import netCDF4
import numpy as np
import xarray
from scipy import interpolate

from .utils import return_if_arts_type, get_arts_typename

__all__ = [
    'GriddedField1',
    'GriddedField2',
    'GriddedField3',
    'GriddedField4',
    'GriddedField5',
    'GriddedField6',
    'griddedfield_from_netcdf',
    'griddedfield_from_xarray',
]


class _GriddedField:
    """:class:`GriddedField` implements the same-named ARTS dataype.

    This class provides the facility of storing gridded data. For this purpose
    the grid-axes as well as the data are stored. GriddedFields can be easily
    written to XML-files as they define a clear datastructure.

    :class:`GriddedField` should not be used directly. Use one of the derived
    types such as :class:`GriddedField1` instead.

    Note:
        For the special case of storing atmospheric profiles as GriddedField3
        the latitude and longitude grids have to be initialised as empty
        np.array.

    Examples:
        Create and manipulate a :class:`GriddedField` object.

        >>> gf1 = GriddedField1()
        >>> gf1.grids = [np.arange(10)]
        >>> gf1.gridnames = ["Indices"]
        >>> gf1.data = np.random.randn(10)

        Inspect an existing :class:`GriddedField` object.

        >>> gf1.dimension
        1
        >>> gf1.grids
        [array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])]
        >>> gf1.gridnames
        ['Indices']

    """

    def __init__(self, dimension=None, grids=None, data=None, gridnames=None,
                 dataname=None, name=None):
        """Create a GriddedField object.

        Parameters:
            dimension (int): Dimension of the GriddedField.
            grids (list, tuple, np.ndarray): grids.
            data (np.ndarray): data values.
            gridnames (List[str]): clear names for all grids.
            dataname (str): name of the data array.
            name (str): name of the GriddedField.

        """
        if not isinstance(dimension, numbers.Integral) or dimension < 1:
            raise ValueError('dimension must be a scalar greater 0')
        self._dimension = dimension
        self.grids = copy.deepcopy(grids)
        self.data = copy.copy(data)
        self.gridnames = copy.copy(gridnames)
        self.dataname = dataname
        self.name = name

    def __getitem__(self, index):
        """Make the data array subscriptable directly.

        ``gf[0, 1]`` is equivalent to ``gf.data[0, 1]``.
        """
        return self.data[index]

    def __eq__(self, other):
        """Test the equality of GriddedFields."""
        if (isinstance(other, self.__class__) or
                isinstance(self, other.__class__)):
            # Check each attribute after another for readabilty.
            # Return as soon as possible for performance.
            if self.name != other.name:
                return False

            if self.dataname != other.dataname:
                return False

            if self.gridnames != other.gridnames:
                return False

            if self.dimension != other.dimension:
                return False

            if self.grids is not None and other.grids is not None:
                if not np.all(a == b for a, b in zip(self.grids, other.grids)):
                    return False
            elif self.grids is not other.grids:
                return False

            if self.data is not None and other.data is not None:
                if not np.allclose(self.data, other.data):
                    return False
            elif self.data is not other.data:
                return False

            return True
        return NotImplemented

    def __neq__(self, other):
        """Test the non-equality of GriddedFields."""
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        try:
            if self.name:
                out = "GriddedField{}: {}\n".format(self.dimension, self.name)
            else:
                out = "GriddedField{}: {}\n".format(self.dimension,
                                                    "Generic Field")

            for i in range(self.dimension):
                if self.gridnames[i]:
                    out += "{} {}: {}\n".format(self.gridnames[i],
                                                np.shape(self.grids[i]),
                                                self.grids[i])
                else:
                    out += "{} {}: {}\n".format("Grid " + str(i + 1),
                                                np.shape(self.grids[i]),
                                                self.grids[i])
            if self.dataname:
                out += "{} {}: {}\n".format(self.dataname,
                                            self.data.shape,
                                            self.data.flatten())
            else:
                out += "{} {}: {}\n".format("Data",
                                            self.data.shape,
                                            self.data.flatten())
            return out
        except:
            # If representation fails, fall back to default.
            # Known issues: Empty GriddedFields.
            return '<{0} at {1}>'.format(type(self).__name__, hex(id(self)))

    @property
    def shape(self):
        """Shape of the data array."""
        return self.data.shape

    @property
    def dimension(self):
        """Dimension of the GriddedField.

        The dimension has to be defined when creating the GriddedField object.
        For the convenience subclasses (e.g. GriddedField1) this is done
        automatically.

        """
        return self._dimension

    @property
    def grids(self):
        """List of grids defining the GriddedField.

        Note:
            The number of grids has to match the GriddedField dimension.
        """
        return self._grids

    @property
    def gridnames(self):
        """A list or tuple that includes a name for every grid.

        Note:
            The number of gridnames has to match the number of grids.
            Gridnames are currently not used so it is not neccesarry
            to set them.
        """
        return self._gridnames

    @property
    def data(self):
        """The data matrix stored in the GriddedField.

        Note:
            The data array has to fit the grid dimensions.
        """
        return self._data

    @property
    def name(self):
        """Name of the GriddedField."""
        return self._name

    @property
    def dataname(self):
        """Name of the data array."""
        return self._dataname

    @grids.setter
    def grids(self, grids):
        if grids is None:
            self._grids = None
            return

        if type(grids) not in (list, tuple):
            raise TypeError('The array of grids must be type list or tuple.')

        for grid in grids:
            if (get_arts_typename(grid)
                    in ['ArrayOfString', 'ArrayOfIndex', 'Vector', None]):
                self._grids = grids
            else:
                raise TypeError(
                    'grids have to be ArrayOfString, ArrayOfIndex or Vector.')

    @gridnames.setter
    def gridnames(self, gridnames):
        self._gridnames = return_if_arts_type(gridnames, 'ArrayOfString')

    @data.setter
    def data(self, data):
        data_type = get_arts_typename(np.ndarray([0] * self.dimension))
        self._data = return_if_arts_type(data, data_type)

    @dataname.setter
    def dataname(self, dataname):
        self._dataname = return_if_arts_type(dataname, 'String')

    @name.setter
    def name(self, name):
        self._name = return_if_arts_type(name, 'String')

    def check_dimension(self):
        """Checks the consistency of grids and data.

        This functions check if the dimensions defined by the grids fit to the
        dimension of the passed data.
        Also check if the number of gridnames fits the number of grids.

        Note:
            This check is done automatically before storing and after loading
            XML files.

        Returns:
            True if successful.

        Raises:
            Exception: if number of grids does not fit
                the GriddedField dimension.
            Exception: if number of gridnames does not fit
                the number of grids.
            Exception: if data dimension does not fit the grid dimensions.
            Warning: if a dimension is empty.

        """
        # define error messages
        grid_dim_error = (('The number of grids has to fit the dimension '
                           'of the GriddedField.\nThe dimension is {0} '
                           'but {1} grids were passed.')
                          .format(self.dimension, len(self.grids)))

        # number of grids has to match the GriddedField dimension
        if len(self.grids) != self.dimension:
            raise Exception(grid_dim_error)

        # if grids are named, each grid has to be named
        if self.gridnames is None:
            self.gridnames = [''] * self.dimension

        grid_name_error = (('The number of gridnames has to fit the '
                            'dimension of the GriddedField.\nThe dimension'
                            ' is {0} but {1} gridnames were passed.')
                           .format(self.dimension, len(self.gridnames)))

        if len(self.gridnames) != self.dimension:
            raise Exception(grid_name_error)

        # grid and data dimension have to fit
        g_dim = [np.size(g) if np.size(g) > 0 else 1 for g in self.grids]

        if tuple(g_dim) != self.data.shape:
            raise Exception(('Dimension mismatch between data and grids. '
                             'Grid dimension is {0} but data {1}')
                            .format(tuple(g_dim), self.data.shape))

        return True

    def check_atm_fields_compact(self, check_gridnames=False):
        """Checks the consistency of grids and data.

        This functions check if the dimensions defined by the grids fit to the
        dimension of the passed data.
        Also check if the number of gridnames fits the number of grids.

        Note:
            The last three gridnames are expected to be 'Pressure', 'Latitude'
            and 'Longitude'. This is more strict than the ARTS checks, but good
            behavior.

        Returns:
            True if successful.

        Raises:
            Exception:
                - If not GriddedField4.
                - If the pressure grid is not decreasing.

        """
        if self.dimension != 4:
            raise Exception('atm_fields_compact have to be GriddedField4.')

        if check_gridnames:
            if self.gridnames[1] != 'Pressure':
                err = "Second grid has to be 'Pressure' not '{}'."
                raise Exception(err.format(self.gridnames[1]))
            elif self.gridnames[2] != 'Latitude':
                err = "Third grid has to be 'Latitude' not '{}'."
                raise Exception(err.format(self.gridnames[2]))
            elif self.gridnames[3] != 'Longitude':
                err = "Fourth grid has to be 'Longitude' not '{}'."
                raise Exception(err.format(self.gridnames[3]))

        if not get_arts_typename(self.grids[0]) == 'ArrayOfString':
            raise Exception('First grid has to be ArrayOfString.')

        if not all(np.diff(self.grids[1]) < 0):
            raise Exception('Pressure grid has to be strictly decreasing.')

        return True

    def copy(self):
        """Return a deepcopy of the GriddedField."""
        return copy.deepcopy(self)

    def extract_slice(self, s=slice(None), axis=0):
        """Return a new GriddedField containing a slice of the current one.

        Parameters:
            s (slice): Slice.
            axis (int): Axis to slice along.

        Returns:
            :class:`typhon.arts.griddedfield._GriddedField`:
                GriddedField containing sliced grids and data.
        """
        gf = self.copy()
        gf.grids[axis] = gf.grids[axis][s]
        slices = [slice(None)] * self.dimension
        slices[axis] = s
        gf.data = gf.data[tuple(slices)]

        return gf

    def refine_grid(self, new_grid, axis=0, fun=np.array, **kwargs):
        """Interpolate GriddedField axis to a new grid.

        This function replaces a grid of a GriddField and interpolates all
        data to match the new coordinates. :func:`scipy.interpolate.interp1d`
        is used for interpolation.

        Parameters:
            new_grid (ndarray): The coordinates of the interpolated values.
            axis (int): Specifies the axis of data along which to interpolate.
                Interpolation defaults to the first axis of the GriddedField.
            fun (numpy.ufunc, or similar): Function to apply to grid before
                interpolation.  Suggested values: np.array, np.log10, np.log
            **kwargs:
                Keyword arguments passed to :func:`scipy.interpolate.interp1d`.

        Returns: :class:`typhon.arts.griddedfield.GriddedField`

        """
        if len(self.grids[axis]) > 1:
            f = interpolate.interp1d(fun(self.grids[axis]), self.data,
                                     axis=axis, **kwargs)
            self.grids[axis] = new_grid
            self.data = f(fun(new_grid))
        else:  # if the intention is to create a useful TensorX
            self.data = self.data.repeat(len(new_grid), axis=axis)
            self.grids[axis] = new_grid

        self.check_dimension()

        return self

    def get(self, key, default=None, keep_dims=True):
        """Return data from field with given fieldname.

        Notes:
              This method only works, if the first grid
              is an :arts:`ArrayOfString`.

        Parameters:
              key (str): Name of the field to extract.
              default: Default value, if ``key`` is not found.
              keep_dims (bool): If ``False``, empty dimensions are squeezed
                  before the extracted array is returned.

        Returns:
            ndarray: Extracted ndarray.
        """
        # The first grid has to be an ArrayOfString.
        if not get_arts_typename(self.grids[0]) == 'ArrayOfString':
            raise TypeError(
                'Method only works, if the first grid is an "ArrayOfString"')

        # If the GriddedField is empty or the given fieldname is not found,
        # return the default value.
        if self.grids is None or key not in self.grids[0]:
            return default

        # Find the index of given fieldname in the name grid and return the
        # ndarray at that position.
        field = self.data[[self.grids[0].index(key)]]

        # Squeeze empty dimensions, if ``keep_dims`` is ``False``.
        return field if keep_dims else field.squeeze()

    def set(self, key, data):
        """Assign data to field with given fieldname.

        Notes:
              This method only works, if the first grid
              is an :arts:`ArrayOfString`.

        Parameters:
              key (str): Name of the field to extract.
              data (ndarray): Data array.
        """
        if not get_arts_typename(self.grids[0]) == 'ArrayOfString':
            raise TypeError(
                'Method only works, if the first grid is an "ArrayOfString"')

        self.data[[self.grids[0].index(key)]] = data

    def scale(self, key, factor, dtype=float):
        """Scale data stored in field with given fieldname.

        Notes:
              This method only works, if the first grid
              is an :arts:`ArrayOfString`.

        Parameters:
              key (str): Name of the field to scale.
              factor (float or ndarray): Scale factor.
              dtype (type): Data type used for typecasting. If the original
                dtype of ``GriddedField.data`` is ``int``, the data array
                gets typecasted to prevent messy behaviour when assigning
                scaled values.
        """
        if issubclass(self.data.dtype.type, numbers.Integral):
            # Typecast integer data arrays to prevent unwanted typecast when
            # assigning scaled (float) variables back to the (integer) ndarray.
            self.data = self.data.astype(dtype)

        self.set(key, self.get(key) * factor)

    def add(self, key, offset, dtype=float):
        """Add offset to data stored in field with given fieldname.

        Notes:
              This method only works, if the first grid
              is an :arts:`ArrayOfString`.

        Parameters:
              key (str): Name of the field to offset.
              offset (float or ndarray): Offset.
              dtype (type): Data type used for typecasting. If the original
                dtype of ``GriddedField.data`` is ``int``, the data array
                gets typecasted to prevent messy behaviour when assigning
                scaled values.
        """
        if issubclass(self.data.dtype.type, numbers.Integral):
            # Typecast integer data arrays to prevent unwanted typecast when
            # assigning scaled (float) variables back to the (integer) ndarray.
            self.data = self.data.astype(dtype)

        self.set(key, self.get(key) + offset)

    def to_dict(self):
        """Convert GriddedField to dictionary.

        Converts a GriddedField object into a classic Python dictionary. The
        gridname is used as dictionary key. If the grid is unnamed the key is
        generated automatically ('grid1', 'grid2', ...). The data can be
        accessed through the 'data' key.

        Returns:
            Dictionary containing the grids and data.

        """
        grids, gridnames = self.grids, self.gridnames

        if gridnames is None:
            gridnames = ['grid%d' % n for n in range(1, self.dimension + 1)]

        for n, name in enumerate(gridnames):
            if name == '':
                gridnames[n] = 'grid%d' % (n + 1)

        d = {name: grid for name, grid in zip(gridnames, grids)}

        if self.dataname is not None:
            d[self.dataname] = self.data
        else:
            d['data'] = self.data

        return d

    def to_xarray(self):
        """Convert GriddedField to xarray.DataArray object.

        Convert a GriddedField object into a :func:`xarray.DataArray`
        object.  The dataname is used as the DataArray name.

        Returns:
            xarray.DataArray object corresponding to gridded field
        """

        da = xarray.DataArray(self.data)
        da = da.rename(dict((k, v)
            for (k, v) in zip(da.dims, self.gridnames)
            if v!=""))
        da = da.assign_coords(
            **{name: coor
                for (name, coor) in zip(da.dims, self.grids)
                if len(coor)>0})
        if self.name is not None:
            da.name = self.name
        da.attrs['data_name'] = self.dataname
        return da

    @classmethod
    def from_nc(cls, inputfile, variable, fill_value=np.nan):
        """Create GriddedField from variable in netCDF files.

        Extract a given variable from a netCDF file. The data and its
        dimensions are returned as a :class:`GriddedField` object.

        Parameters:
            inputfile (str): Path to netCDF file.
            variable (str): Variable key of variable to extract.
            fill_value (float): Fill value for masked areas (default: np.nan).

        Returns:
            GriddedField object of sufficient dimension.

        Raises:
            Exception: If the variable key can't be found in the netCDF file.

        """
        with netCDF4.Dataset(inputfile) as nc:
            if variable not in nc.variables:
                raise Exception('netCDF file has no variable {}.'.format(variable))

            data = nc.variables[variable]

            obj = cls()
            obj.grids = [nc.variables[dim][:] for dim in data.dimensions]
            obj.gridnames = [dim for dim in data.dimensions]

            if isinstance(data[:], np.ma.MaskedArray):
                obj.data = data[:].filled(fill_value=fill_value)
            else:
                obj.data = data[:]

        obj.check_dimension()

        return obj

    @classmethod
    def from_xml(cls, xmlelement):
        """Load a GriddedField from an ARTS XML file.

        Returns:
            GriddedField. Dimension depends on data in file.


        """
        obj = cls()

        if 'name' in xmlelement.attrib:
            obj.name = xmlelement.attrib['name']

        obj.grids = [x.value() for x in xmlelement[:-1]]
        obj.gridnames = [x.attrib['name']
                         if 'name' in x.attrib else ''
                         for x in xmlelement[:-1]]

        # Read data (and optional dataname).
        obj.data = xmlelement[-1].value()
        if 'name' in xmlelement[-1].attrib:
            obj.dataname = xmlelement[-1].attrib['name']

        obj.check_dimension()
        return obj

    @classmethod
    def from_xarray(cls, da):
        """Create GriddedField from a xarray.DataArray object.

        The data and its dimensions are returned as a :class:`GriddedField` object.
        The DataArray name is used as name for the gridded field. If the attribute
        `data_name` is present, it is used as `dataname` on the :class:`GriddedField`.

        Parameters:
            da (xarray.DataArray): xarray.DataArray containing the dimensions and data.

        Returns:
            GriddedField object.

        """
        obj = cls()
        obj.grids = [da[c].values for c in da.dims]
        obj.gridnames = list(da.dims)
        obj.data = da.values
        obj.dataname = da.attrs.get('data_name', 'Data')
        if da.name:
            obj.name = da.name
        obj.check_dimension()
        return obj

    def to_atmlab_dict(self):
        """Returns a copy of the GriddedField as a dictionary.

        Returns a dictionary compatible with an atmlab structure.

        Returns:
            Dictionary containing the grids and data.
        """

        d = {}
        if self.name is None:
            d['name'] = ''
        else:
            d['name'] = self.name
        d['grids'] = self.grids
        d['gridnames'] = self.gridnames
        d['data'] = self.data
        if self.dataname is None:
            d['dataname'] = ''
        else:
            d['dataname'] = self.dataname

        return d

    def write_xml(self, xmlwriter, attr=None):
        """Save a GriddedField to an ARTS XML file."""
        self.check_dimension()

        if attr is None:
            attr = {}

        if self.name is not None:
            attr['name'] = self.name

        xmlwriter.open_tag('GriddedField{}'.format(self.dimension), attr)
        for grid, name in zip(self.grids, self.gridnames):
            xmlwriter.write_xml(grid, {'name': name})

        if self.dataname is None:
            xmlwriter.write_xml(self.data)
        else:
            xmlwriter.write_xml(self.data, {'name': self.dataname})

        xmlwriter.close_tag()


class GriddedField1(_GriddedField):
    """Implements a :arts:`GriddedField1`."""

    def __init__(self, *args, **kwargs):
        super(GriddedField1, self).__init__(1, *args, **kwargs)


class GriddedField2(_GriddedField):
    """Implements a :arts:`GriddedField2`."""

    def __init__(self, *args, **kwargs):
        super(GriddedField2, self).__init__(2, *args, **kwargs)


class GriddedField3(_GriddedField):
    """Implements a :arts:`GriddedField3`."""

    def __init__(self, *args, **kwargs):
        super(GriddedField3, self).__init__(3, *args, **kwargs)


class GriddedField4(_GriddedField):
    """Implements a :arts:`GriddedField4`."""

    def __init__(self, *args, **kwargs):
        super(GriddedField4, self).__init__(4, *args, **kwargs)


class GriddedField5(_GriddedField):
    """Implements a :arts:`GriddedField5`."""

    def __init__(self, *args, **kwargs):
        super(GriddedField5, self).__init__(5, *args, **kwargs)


class GriddedField6(_GriddedField):
    """Implements a :arts:`GriddedField6`."""

    def __init__(self, *args, **kwargs):
        super(GriddedField6, self).__init__(6, *args, **kwargs)


def _griddedfield_from_ndim(ndim):
    """Determine proper GriddedField type from number of dimensions."""
    griddefield_dimension_map = {
        1: GriddedField1,
        2: GriddedField2,
        3: GriddedField3,
        4: GriddedField4,
        5: GriddedField5,
        6: GriddedField6,
    }
    return griddefield_dimension_map[ndim]


def griddedfield_from_netcdf(ncfile, **kwargs):
    """Create an ARTS ``GriddedField`` of appropriate dimension from netCDF.

    Parameters:
        ncfile (str): Path to netCDF file.
        **kwargs: Additional keyword arguments are passed
            to the :func:`~GriddedField1.from_nc` method.

    Returns:
        : Appropriate ARTS GriddedField.
    """
    with netCDF4.Dataset(ncfile) as root:
        cls = _griddedfield_from_ndim(root.ndim)

    return cls.from_nc(ncfile, **kwargs)


def griddedfield_from_xarray(dataarray):
    """Convert :class:`xarray.DataArray` to ARTS ``GriddedField``.

    Parameters:
        dataarray (:class:`xarray.DataArray`): :class:`~xarray.DataArray`
            containing dimensions and data.

    Returns:
        : Appropriate ARTS GriddedField.
    """
    return _griddedfield_from_ndim(dataarray.ndim).from_xarray(dataarray)



"""Retrieval of IWP from passive radiometers

This class is a reimplementation of the SPARE-ICE product introduced by
Holl et al. 2014.

References:
    TODO: Add reference.

Examples:

    .. code-block:: python

    from typhon.files import AVHRR_GAC_HDF, CloudSat, FileSet, MHS_HDF
    from typhon.retrieval import SPAREICE

    cloudsat = FileSet(...)
    mhs = FileSet(...)
    avhrr = FileSet(...)

    spareice = SPAREICE(
        file="spareice.json",
    )

    # Either we have already collocated, then we can use the files directly for
    # the training or SPARE-ICE should create the training dataset by itself
    # (should collocate by itself).
    data = spareice.prepare_training_data(
        # Do we have already collocations with all instruments? Put them here:
        collocations=...,
        # OR
        cloudsat=cloudsat, mhs=mhs, avhrr=avhrr,
        # Which time period should be used for training?
        start=..., end=...,
    )

    # To save time and memory space, we can store the current object with
    # the training data to the disk and reuse it later directly. So, we do not
    # have to call spareice.prepare_training_data again:
    data.to_netcdf("spareice_training.nc")

    # Train SPARE-ICE with the data
    spareice.train(data, test_ratio=0.2)

    # After training, we can use the SPARE-ICE retrieval:
    spareice.retrieve(
        # Do we have already collocations with MHS and AVHRR? Put them here:
        collocations=...,
        # Otherwise we can put here each fileset and create collocations
        # on-the-fly
        mhs=mhs, avhrr=avhrr,
        # Which time period should be used for retrieving?
        start=..., end=...,
        output=...,
    )
"""
from ast import literal_eval
import itertools
import logging
import os
from os.path import join, dirname
import warnings

import imageio
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import binned_statistic as sci_binned_statistic
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import GridSearchCV
from sklearn.neural_network import MLPRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import RobustScaler
from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier
from typhon.collocations import collapse, Collocations, Collocator
from typhon.geographical import sea_mask
from typhon.plots import binned_statistic, heatmap, styles, worldmap
from typhon.utils import to_array, Timer
import xarray as xr

from ..common import RetrievalProduct

__all__ = [
    'SPAREICE',
]

# The path to the standard weights:
PARAMETERS_DIR = join(dirname(__file__), 'parameters')
STANDARD_FILE = join(PARAMETERS_DIR, "standard.json")


logger = logging.getLogger(__name__)


class SPAREICE:
    """Retrieval of IWP from passive radiometers

    Examples:
    .. code-block:: python

        import pandas as pd
        from typhon.retrieval import SPAREICE

        # Create a SPARE-ICE object with the standard weights
        spareice = SPAREICE()

        # Print the required input fields
        print(spareice.inputs)

        # If you want to know the input fields for the each component, IWP
        # regressor and ice cloud classifier, you can get them like this:
        print(spareice.iwp.inputs)  # Inputs from IWP regressor
        print(spareice.ice_cloud.inputs)  # Inputs from ice cloud classifier

        # If you have yur own input data, you can use :meth:`retrieve` to run
        # SPARE-ICE on it.
        data = pd.DataFrame(...)
        retrieved = spareice.retrieve(data)

        # If your data directly comes from collocations between MHS and AVHRR,
        # you can use ::meth:`convert_collocated_data` to make it SPARE-ICE
        # compatible.
        collocations = Collocator().collocate(mhs_data, avhrr_data, ...)
        standardized_data = self.standardize_collocations(collocations)
        retrieved = spareice.retrieve(standardized_data)
    """

    def __init__(self, file=None, collocator=None, processes=10, verbose=0,
                 sea_mask_file=None, elevation_file=None):
        """Initialize a SPAREICE object

        Args:
            file: A JSON file with the coefficients of SPAREICE. If not given,
                the standard configuration will be loaded.
            collocator: SPARE-ICE requires a collocator when it should be
                generated from filesets. You can pass your own
                :class:`Collocator` object here if you want.
            processes: Number of processes to parallelize the training or
                collocation search. 10 is the default. Best value depends on
                your machine.
            verbose (int): Control ``GridSearchCV`` verbosity. The higher the
            value, the more debug messages are printed.
        """

        self.verbose = verbose
        self.processes = processes
        self.name = "SPARE-ICE"

        if sea_mask_file is None:
            self.sea_mask = None
        else:
            self.sea_mask = np.flip(
                np.array(imageio.imread(sea_mask_file) == 255), axis=0
            )

        if elevation_file is None:
            self.elevation_grid = None
        else:
            ds = xr.open_dataset(
                elevation_file, decode_times=False
            )
            self.elevation_grid = ds.data.squeeze().values

        if collocator is None:
            self.collocator = Collocator()
        else:
            self.collocator = collocator

        # SPARE-ICE consists of two retrievals: one neural network for the IWP
        # and one decision tree classifier for the ice cloud flag
        self._iwp = None
        self._ice_cloud = None

        # The users can load SPARE-ICE from their own training or the standard
        # parameters:
        if file is None:
            try:
                self.load(STANDARD_FILE)
            except Exception as e:
                warnings.warn(
                    "Could not load the standard parameters of SPARE-ICE!\n"
                    "You need to train SPARE-ICE by yourself."
                )
                warnings.warn(str(e))
                self._iwp = RetrievalProduct()
                self._ice_cloud = RetrievalProduct()
        else:
            self.load(file)

    def _debug(self, msg):
        logger.debug(f"[{self.name}] {msg}")

    def _info(self, msg):
        logger.info(f"[{self.name}] {msg}")

    def _iwp_model(self, processes, cv_folds):
        """Return the default model for the IWP regressor
        """
        # Estimators are normally objects that have a fit and predict method
        # (e.g. MLPRegressor from sklearn). To make their training easier we
        # scale the input data in advance. With Pipeline objects from sklearn
        # we can combine such steps easily since they behave like an
        # estimator object as well.
        estimator = Pipeline([
            # SVM or NN work better if we have scaled the data in the first
            # place. MinMaxScaler is the simplest one. RobustScaler or
            # StandardScaler could be an alternative.
            ("scaler", RobustScaler(quantile_range=(15, 85))),
            # The "real" estimator:
            ("estimator", MLPRegressor(max_iter=6000, early_stopping=True)),
        ])

        # To optimize the results, we try different hyper parameters by
        # using a grid search
        hidden_layer_sizes = [
            (15, 10, 3),
            #(50, 20),
        ]
        hyper_parameter = [
            {   # Hyper parameter for lbfgs solver
                'estimator__solver': ['lbfgs'],
                'estimator__activation': ['tanh'],
                'estimator__hidden_layer_sizes': hidden_layer_sizes,
                'estimator__random_state': [0, 42, 100, 3452],
                'estimator__alpha': [0.1, 0.001, 0.0001],
            },
        ]

        return GridSearchCV(
            estimator, hyper_parameter, refit=True,
            n_jobs=processes, cv=cv_folds, verbose=self.verbose,
        )

    @staticmethod
    def _ice_cloud_model():
        """Return the default model for the ice cloud classifier"""
        # As simple as it is. We do not need a grid search trainer for the DTC
        # since it has already a good performance.
        return DecisionTreeClassifier(
            max_depth=12, random_state=5, # n_estimators=20, max_features=9,
        )

    @property
    def inputs(self):
        """Return the input fields of the current configuration"""
        return list(set(self.iwp.inputs) | set(self.ice_cloud.inputs))

    @property
    def iwp(self):
        """Return the IWP regressor of SPARE-ICE"""
        return self._iwp

    @property
    def ice_cloud(self):
        """Return the ice cloud classifier of SPARE-ICE"""
        return self._ice_cloud

    def load(self, filename):
        """Load SPARE-ICE from a json file

        Args:
            filename: Path and name of the file.

        Returns:
            None
        """
        with open(filename, 'r') as infile:
            parameters = literal_eval(infile.read())
            self._iwp = RetrievalProduct.from_dict(
                parameters["iwp"]
            )
            self._ice_cloud = RetrievalProduct.from_dict(
                parameters["ice_cloud"]
            )

    def save(self, filename):
        """Save SPARE-ICE to a json file

        Notes:
            The output format is not standard json!

        Args:
            filename: Path and name of the file.

        Returns:
            None
        """
        with open(filename, 'w') as outfile:
            dictionary = {
                "iwp": self.iwp.to_dict(),
                "ice_cloud": self.ice_cloud.to_dict(),
            }
            outfile.write(repr(dictionary))

    def standardize_collocations(self, data, fields=None, add_sea_mask=True,
                                 add_elevation=True):
        """Convert collocation fields to standard SPARE-ICE fields.

        Args:
            data: A xarray.Dataset object with collocations either amongst
                2C-ICE, MHS & AVHRR or MHS & AVHRR.
            fields (optional): Fields that will be selected from the
                collocations. If None (default), all fields will be selected.
            add_sea_mask: Add a flag to the data whether the pixel is over sea
                or land.
            add_elevation: Add the surface elevation in meters to each pixel.

        Returns:
            A pandas.DataFrame with all selected fields.
        """
        # Check whether the data is coming from a twice-collocated dataset:
        if "MHS_2C-ICE/MHS/scnpos" in data.variables:
            prefix = "MHS_2C-ICE/"
        else:
            prefix = ""

        # The keys of this dictionary are the new names, while the values are
        # old the names coming from the original collocations. If the value is
        # a list, the variable is 2-dimensional. The first element is the old
        # name, and the rest is the dimnesion that should be selected.
        mapping = {
            "mhs_channel1": [
                f"{prefix}MHS/Data/btemps", f"{prefix}MHS/channel", 0
            ],
            "mhs_channel2": [
                f"{prefix}MHS/Data/btemps", f"{prefix}MHS/channel", 1
            ],
            "mhs_channel3": [
                f"{prefix}MHS/Data/btemps", f"{prefix}MHS/channel", 2
            ],
            "mhs_channel4": [
                f"{prefix}MHS/Data/btemps", f"{prefix}MHS/channel", 3
            ],
            "mhs_channel5": [
                f"{prefix}MHS/Data/btemps", f"{prefix}MHS/channel", 4
            ],
            "lat": "lat",
            "lon": "lon",
            "time": "time",
            "mhs_scnpos": f"{prefix}MHS/scnpos",
            "solar_azimuth_angle":
                f"{prefix}MHS/Geolocation/Solar_azimuth_angle",
            "solar_zenith_angle":
                f"{prefix}MHS/Geolocation/Solar_zenith_angle",
            "satellite_azimuth_angle":
                f"{prefix}MHS/Geolocation/Satellite_azimuth_angle",
            "satellite_zenith_angle":
                f"{prefix}MHS/Geolocation/Satellite_zenith_angle",
            "avhrr_channel1": [
                "AVHRR/Data/btemps_mean", "AVHRR/channel", 0
            ],
            "avhrr_channel2": [
                "AVHRR/Data/btemps_mean", "AVHRR/channel", 1
            ],
            "avhrr_channel3": [
                "AVHRR/Data/btemps_mean", "AVHRR/channel", 2
            ],
            "avhrr_channel4": [
                "AVHRR/Data/btemps_mean", "AVHRR/channel", 3
            ],
            "avhrr_channel5": [
                "AVHRR/Data/btemps_mean", "AVHRR/channel", 4
            ],
            "avhrr_channel1_std": [
                "AVHRR/Data/btemps_std", "AVHRR/channel", 0
            ],
            "avhrr_channel2_std": [
                "AVHRR/Data/btemps_std", "AVHRR/channel", 1
            ],
            "avhrr_channel3_std": [
                "AVHRR/Data/btemps_std", "AVHRR/channel", 2
            ],
            "avhrr_channel4_std": [
                "AVHRR/Data/btemps_std", "AVHRR/channel", 3
            ],
            "avhrr_channel5_std": [
                "AVHRR/Data/btemps_std", "AVHRR/channel", 4
            ],
            "iwp_number": "MHS_2C-ICE/2C-ICE/ice_water_path_number",
            "iwp_std": "MHS_2C-ICE/2C-ICE/ice_water_path_std",
        }

        # These fields need a special treatment
        special_fields = ["avhrr_tir_diff", "mhs_diff", "iwp", "ice_cloud"]

        # Default - take all fields:
        if fields is None:
            fields = list(mapping.keys()) + special_fields

        return_data = {}
        for field in fields:
            if field in special_fields:
                # We will do this later:
                continue
            elif field not in mapping:
                # Some fields might be added later (such as elevation, etc)
                continue

            key = mapping[field]
            try:
                if isinstance(key, list):
                    return_data[field] = data[key[0]].isel(
                        **{key[1]: key[2]}
                    )
                else:
                    return_data[field] = data[key]
            except KeyError:
                # Keep things easy. Collocations might contain the target
                # dataset or not. We do not want to have a problem just because
                # we have not them.
                pass

        return_data = pd.DataFrame(return_data)

        if "avhrr_tir_diff" in fields:
            return_data["avhrr_tir_diff"] = \
                return_data["avhrr_channel5"] - return_data["avhrr_channel4"]
        if "mhs_diff" in fields:
            return_data["mhs_diff"] = \
                return_data["mhs_channel5"] - return_data["mhs_channel3"]
        if "iwp" in fields and "MHS_2C-ICE/2C-ICE/ice_water_path_mean" in data:
            # We transform the IWP to log space because it is better for the
            # ANN training. Zero values might trigger warnings and
            # result in -INF. However, we cannot drop them because the ice
            # cloud classifier needs zero values for its training.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return_data["iwp"] = np.log10(
                    data["MHS_2C-ICE/2C-ICE/ice_water_path_mean"]
                )
                return_data["iwp"].replace(
                    [-np.inf, np.inf], np.nan, inplace=True
                )
        if "ice_cloud" in fields \
                and "MHS_2C-ICE/2C-ICE/ice_water_path_mean" in data:
            return_data["ice_cloud"] = \
                data["MHS_2C-ICE/2C-ICE/ice_water_path_mean"] > 0

        if add_sea_mask:
            return_data["sea_mask"] = sea_mask(
                return_data.lat, return_data.lon, self.sea_mask
            )

        if add_elevation:
            def get_grid_value(grid, lat, lon):
                lat = to_array(lat)
                lon = to_array(lon)

                if lon.min() < -180 or lon.max() > 180:
                    raise ValueError("Longitudes out of bounds!")

                if lat.min() < -90 or lat.max() > 90:
                    raise ValueError("Latitudes out of bounds!")

                grid_lat_step = 180 / (grid.shape[0] - 1)
                grid_lon_step = 360 / (grid.shape[1] - 1)

                lat_cell = (90 - lat) / grid_lat_step
                lon_cell = lon / grid_lon_step

                return grid[lat_cell.astype(int), lon_cell.astype(int)]

            return_data["elevation"] = get_grid_value(
                self.elevation_grid, return_data.lat, return_data.lon
            )

            # We do not need the depth of the oceans (this would just
            # confuse the ANN):
            return_data["elevation"][return_data.elevation < 0] = 0

        return return_data

    def retrieve(self, data, as_log10=False):
        """Retrieve SPARE-ICE for the input variables

        Args:
            data: A pandas.DataFrame object with required input fields (see
                above) or a xarray.Dataset if `from_collocations` is True.
            as_log10: If true, the retrieved IWP will be returned as logarithm
                of base 10.

        Returns:
            A pandas DataFrame object with the retrieved IWP and ice cloud
            flag.
        """

        # Retrieve the ice water path:
        retrieved = self.iwp.retrieve(data[self.iwp.inputs])
        if not as_log10 and retrieved is not None:
            retrieved["iwp"] = 10**retrieved["iwp"]

        # Retrieve the ice cloud flag:
        retrieved = retrieved.join(
            self.ice_cloud.retrieve(data[self.ice_cloud.inputs]),

        )

        return retrieved

    @staticmethod
    def _retrieve_from_collocations(collocations, _, spareice):

        # We need collapsed collocations:
        if "Collocations/pairs" in collocations.variables:
            collocations = collapse(collocations, reference="MHS")

        # However, we do not need the original field names
        collocations = spareice.standardize_collocations(collocations)

        # Remove NaNs from the data:
        collocations = collocations.dropna()

        if collocations.empty:
            return None

        # Retrieve the IWP and the ice cloud flag:
        retrieved = spareice.retrieve(collocations).to_xarray()

        start = collocations.time.min()
        end = collocations.time.max()
        spareice._debug(f"Retrieve SPARE-ICE from {start} to {end}")

        # Add more information:
        retrieved["iwp"].attrs = {
            "units": "g/m^2",
            "name": "Ice Water Path",
            "description": "Ice Water Path (retrieved by SPARE-ICE)."
        }
        retrieved["ice_cloud"].attrs = {
            "units": "boolean",
            "name": "Ice Cloud Flag",
            "description": "True if pixel contains an ice cloud (retrieved"
                           " by SPARE-ICE)."
        }
        retrieved["lat"] = collocations["lat"]
        retrieved["lon"] = collocations["lon"]
        retrieved["time"] = collocations["time"]
        retrieved["scnpos"] = collocations["mhs_scnpos"]

        return retrieved

    def retrieve_from_collocations(
            self, inputs, output, start=None, end=None, processes=None
    ):
        """Retrieve SPARE-ICE from collocations between MHS and AVHRR

        You can use this either with already collocated MHS and AVHRR data
        (pass the :class:`Collocations` object via `inputs`) or you let MHS and
        AVHRR be collocated on-the-fly by passing the filesets with the raw
        data (pass two filesets as list via `inputs`).

        Args:
            inputs: Can be :class:`Collocations` or a list with
                :class:`~typhon.files.fileset.FileSet` objects. If it is a
                :class:`Collocations` object, all files from them are processed
                and use as input for SPARE-ICE.
            output: Must be a path with placeholders or a :class:`FileSet`
                object where the output files should be stored.
            start: Start date either as datetime object or as string
                ("YYYY-MM-DD hh:mm:ss"). Year, month and day are required.
                Hours, minutes and seconds are optional. If not given, it is
                datetime.min per default.
            end: End date. Same format as "start". If not given, it is
                datetime.max per default.
            processes: Number of processes to parallelize the collocation
                search. If not set, the value from the initialization is
                taken.

        Returns:
            None
        """
        if processes is None:
            processes = self.processes

        if "sea_mask" in self.inputs and self.sea_mask is None:
            raise ValueError("You have to pass a sea_mask file via init!")
        if "elevation" in self.inputs and self.elevation_grid is None:
            raise ValueError("You have to pass a elevation file via init!")

        timer = Timer.start()
        if isinstance(inputs, Collocations):
            # Simply apply a map function to all files from these collocations
            inputs.map(
                SPAREICE._retrieve_from_collocations, kwargs={
                    "spareice": self,
                }, on_content=True, pass_info=True, start=start, end=end,
                max_workers=processes, output=output, worker_type="process"
            )
        elif len(inputs) == 2:
            # Collocate MHS and AVHRR on-the-fly:

            names = set(fileset.name for fileset in inputs)
            if "MHS" not in names or "AVHRR" not in names:
                raise ValueError(
                    "You must name the input filesets MHS and AVHRR! Their "
                    f"current names are: {names}"
                )

            iterator = self.collocator.collocate_filesets(
                inputs, start=start, end=end, processes=processes,
                max_interval="30s", max_distance="7.5 km", output=output,
                post_processor=SPAREICE._retrieve_from_collocations,
                post_processor_kwargs={
                    "spareice": self,
                },
            )
            for filename in iterator:
                if filename is not None:
                    self._info(f"Stored SPARE-ICE to\n{filename}")
        else:
            raise ValueError(
                "You need to pass a Collocations object or a list with a MHS "
                "and AVHRR fileset!"
            )
        logger.info(f"Took {timer} hours to retrieve SPARE-ICE")

    def score(self, data):
        """Calculate the score of SPARE-ICE on testing data

        Args:
            data: A pandas.DataFrame object with the required input fields.

        Returns:
            The score for the IWP regressor and the score for the ice cloud
            classifier.
        """

        ice_cloud_score = self.ice_cloud.score(
            data[self.ice_cloud.inputs], data[self.ice_cloud.outputs]
        )

        # We cannot allow NaN or Inf (resulting from transformation to
        # log space)
        data = data.dropna()
        iwp_score = self.iwp.score(
            data[self.iwp.inputs], data[self.iwp.outputs]
        )
        return iwp_score, ice_cloud_score

    def train(self, data, iwp_inputs=None, ice_cloud_inputs=None,
              iwp_model=None, ice_cloud_model=None, processes=None,
              cv_folds=None):
        """Train SPARE-ICE with data

        This trains the IWP regressor and ice cloud classifier.

        Args:
            data: A pandas.DataFrame object with the required input fields.
            iwp_inputs: A list with the input field names for the IWP
                regressor. If this is None, the IWP regressor won't be trained.
            ice_cloud_inputs: A list with the input field names for the ice
                cloud classifier. If this is None, the ice cloud classifier
                won't be trained.
            iwp_model: Set this to your own sklearn estimator class.
            ice_cloud_model: Set this to your own sklearn estimator class.
            processes: Number of processes to parallelize the regressor
                training. If not set, the value from the initialization is
                taken.
            cv_folds: Number of folds used for cross-validation. Default is 5.
                The higher the number the more data is used for training but
                the runtime increases. Good values are between 3 and 10.

        Returns:
            None
        """

        if iwp_inputs is None and ice_cloud_inputs is None:
            raise ValueError("Either fields for the IWP regressor or ice "
                             "cloud classifier must be given!")

        if ice_cloud_inputs is not None:
            self._info("Train SPARE-ICE - ice cloud classifier")
            if ice_cloud_model is None:
                ice_cloud_model = self._ice_cloud_model()

            score = self.ice_cloud.train(
                ice_cloud_model,
                data[ice_cloud_inputs], data[["ice_cloud"]],
            )
            self._info(f"Ice cloud classifier training score: {score:.2f}")

        if iwp_inputs is not None:
            self._info("Train SPARE-ICE - IWP regressor")

            # We cannot allow NaN or Inf (resulting from transformation to
            # log space)
            data = data.dropna()

            if processes is None:
                processes = self.processes
            if cv_folds is None:
                cv_folds = 5

            if iwp_model is None:
                iwp_model = self._iwp_model(processes, cv_folds)

            score = self.iwp.train(
                iwp_model, data[iwp_inputs], data[["iwp"]],
            )
            self._info(f"IWP regressor training score: {score:.2f}")
            self._training_report(iwp_model)

    @staticmethod
    def _training_report(trainer):
        if not hasattr(trainer, "cv_results_"):
            return

        logger.info("Best parameters found on training dataset:\n",
              trainer.best_params_)

        means = trainer.cv_results_['mean_test_score']
        stds = trainer.cv_results_['std_test_score']
        for mean, std, params in zip(means, stds, trainer.cv_results_['params']):  # noqa
            logger.info("%0.3f (+/-%0.03f) for %r" % (mean, std * 2, params))

    def report(self, output_dir, experiment, data):
        """Test the performance of SPARE-ICE and plot it

        Args:
            output_dir: A path to a directory (does not need to exist). A
                subdirectory named `experiment` will be created there. All
                plots are stored to it.
            experiment: A name for the experiment as a string. Will be included
                in the title of the plots and used as name for the subdirectory
                in `output_dir`.
            data: A pandas.DataFrame object with the required input fields.

        Returns:
            None
        """
        # Create the output directory:
        output_dir = join(output_dir, experiment)
        os.makedirs(output_dir, exist_ok=True)

        # Run SPARE-ICE!
        retrieved = self.retrieve(data, as_log10=True)

        # We are going to plot the performance of the two retrievals:
        self._report_iwp(output_dir, experiment, data, retrieved)
        self._report_ice_cloud(output_dir, experiment, data, retrieved)

    def _report_iwp(self, output_dir, experiment, test, retrieved):
        """Create and store the plots for IWP regressor"""

        # Plot the heatmap with the retrieved IWPs
        fig, ax = plt.subplots(figsize=(10, 8))
        scat = heatmap(
            test.iwp,
            retrieved.iwp,
            bins=50, range=[[-1, 4], [-1, 4]],
            cmap="density", vmin=5,
        )
        scat.cmap.set_under("w")
        ax.set_xlabel("log10 IWP (2C-ICE) [g/m^2]")
        ax.set_ylabel("log10 IWP (SPARE-ICE) [g/m^2]")
        ax.set_title(experiment)
        fig.colorbar(scat, label="Number of points")
        fig.savefig(join(output_dir, "2C-ICE-SPAREICE_heatmap.png"))

        self._plot_scatter(
            experiment,
            join(output_dir, "2C-ICE-SPAREICE_scatter_{area}.png"),
            test.iwp, retrieved.iwp, test.sea_mask.values
        )

        # MFE plot with 2C-ICE on x-axis
        fe = 100 * (
                np.exp(np.abs(
                    np.log(
                        10 ** retrieved.iwp.values
                        / 10 ** test.iwp.values
                    )
                ))
                - 1
        )
        self._plot_error(
            experiment, join(output_dir, "2C-ICE-SPAREICE_mfe.png"),
            test,
            fe, test.sea_mask.values,
        )

        # Plot the bias:
        bias = retrieved.iwp.values - test.iwp.values
        self._plot_error(
            experiment, join(output_dir, "2C-ICE-SPAREICE_bias.png"),
            test,
            bias, test.sea_mask.values,
            mfe=False, yrange=[-0.35, 0.45],
        )

        # self._plot_weights(
        #     experiment, join(output_dir, "SPAREICE_iwp_weights.png"),
        # )

        with open(join(output_dir, "mfe.txt"), "w") as file:
            mfe = sci_binned_statistic(
                test.iwp.values,
                fe, statistic="median", bins=20,
                range=[0, 4],
            )
            file.write(repr(mfe[0]))

    @staticmethod
    def _plot_scatter(experiment, file, xdata, ydata, sea_mask):
        for area in ["all", "land", "sea"]:
            if area == "all":
                mask = slice(None, None, None)
            elif area == "land":
                mask = ~sea_mask
            else:
                mask = sea_mask

            fig, ax = plt.subplots(figsize=(10, 8))
            ax.scatter(
                xdata[mask], ydata[mask],
                s=1, alpha=0.6
            )
            ax.grid()
            ax.set_xlabel("log10 IWP (2C-ICE) [g/m^2]")
            ax.set_ylabel("log10 IWP (SPARE-ICE) [g/m^2]")
            ax.set_title(f"{experiment} - {area}")
            fig.savefig(file.format(area=area))

    @staticmethod
    def _plot_error(
            experiment, file, xdata, error, sea_mask, mfe=True, yrange=None
    ):

        fig, ax = plt.subplots(figsize=(10, 8))
        xlabel = "log10 IWP (2C-ICE) [g/m^2]"
        xrange = [0, 4]

        if mfe:
            ax.set_ylabel("Median fractional error [%]")
            ax.set_ylim([0, 200])
            statistic = "median"
        else:
            ax.set_ylabel("$\Delta$ IWP (SPARE-ICE - 2C-ICE) [log 10 g/m^2]")
            statistic = "mean"

        for hemisphere in ["global"]:
            for area in ["all", "land", "sea"]:
                if area == "all":
                    mask = np.repeat(True, xdata.iwp.size)
                elif area == "land":
                    mask = ~sea_mask
                else:
                    mask = sea_mask

                if hemisphere == "north":
                    mask &= xdata.lat.values >= 0
                elif hemisphere == "south":
                    mask &= xdata.lat.values < 0

                binned_statistic(
                    xdata.iwp.values[mask],
                    error[mask], statistic=statistic, bins=20,
                    range=xrange, pargs={"marker": "o",
                                         "label": f"{area} - {hemisphere}"}
                )

        ax.set_xlabel(xlabel)
        ax.grid()
        ax.legend(fancybox=True)
        ax.set_title(f"Experiment: {experiment}")
        if yrange is not None:
            ax.set_ylim(yrange)
        fig.tight_layout()
        fig.savefig(file)

    def _plot_weights(self, title, file, layer_index=0, vmin=-5, vmax=5):
        import seaborn as sns
        sns.set_context("paper")

        layers = self.iwp.estimator.steps[-1][1].coefs_
        layer = layers[layer_index]
        f, ax = plt.subplots(figsize=(18, 12))
        weights = pd.DataFrame(layer)
        weights.index = self.iwp.inputs

        sns.set(font_scale=1.1)

        # Draw a heatmap with the numeric values in each cell
        sns.heatmap(
            weights, annot=True, fmt=".1f", linewidths=.5, ax=ax,
            cmap="difference", center=0, vmin=vmin, vmax=vmax,
            # annot_kws={"size":14},
        )
        ax.tick_params(labelsize=18)
        f.tight_layout()
        f.savefig(file)

    def _report_ice_cloud(self, output_dir, experiment, test, retrieved):
        # Confusion matrix:
        fig, ax = plt.subplots(figsize=(12, 10))
        cm = confusion_matrix(test.ice_cloud, retrieved.ice_cloud)
        img = self._plot_matrix(cm, classes=["Yes", "No"], normalize=True)
        fig.colorbar(img, label="probability")
        ax.set_title("Ice Cloud Classifier - Performance")
        ax.set_ylabel('real ice cloud')
        ax.set_xlabel('predicted ice cloud')
        fig.tight_layout()
        fig.savefig(join(output_dir, "ice-cloud-confusion-matrix.png"))

        fig, ax = plt.subplots(figsize=(12, 10))
        ax.barh(
            np.arange(len(self.ice_cloud.inputs)),
            self.ice_cloud.estimator.feature_importances_
        )
        ax.set_yticks(np.arange(len(self.ice_cloud.inputs)))
        ax.set_yticklabels(self.ice_cloud.inputs)
        ax.set_xlabel("Feature Importance")
        ax.set_ylabel("Feature")
        ax.set_title("Ice Cloud Classifier - Importance")
        fig.savefig(join(output_dir, "ice-cloud-feature-importance.png"))

    @staticmethod
    def _plot_matrix(
            matrix, classes, normalize=False, ax=None, **kwargs
    ):
        """Plots the confusion matrix of
        Normalization can be applied by setting `normalize=True`.
        """
        if normalize:
            matrix = matrix.astype('float') / matrix.sum(axis=1)[:, np.newaxis]

        default_kwargs = {
            "cmap": "Blues",
            **kwargs
        }

        if ax is None:
            ax = plt.gca()

        img = ax.imshow(matrix, interpolation='nearest', **default_kwargs)
        tick_marks = np.arange(len(classes))
        ax.set_xticks(tick_marks)
        ax.set_xticklabels(classes, rotation=45)
        ax.set_yticks(tick_marks)
        ax.set_yticklabels(classes)

        fmt = '.2f' if normalize else 'd'
        thresh = matrix.max() / 2.
        for i, j in itertools.product(range(matrix.shape[0]),
                                      range(matrix.shape[1])):
            ax.text(j, i, format(matrix[i, j], fmt),
                    horizontalalignment="center",
                    color="white" if matrix[i, j] > thresh else "black")

        return img



# -*- coding: utf-8 -*-
""" Common variables to use within the units subpackage.
"""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

from pint import (UnitRegistry, Context)

__all__ = [
    'ureg',
    'radiance_units',
        ]

ureg = UnitRegistry()
ureg.define("micro- = 1e-6 = µ-")

# aid conversion between different radiance units
sp2 = Context("radiance")
# specrad per frequency → specrad per wavenumber
sp2.add_transformation(
    "[length] * [mass] / [time] ** 3",
    "[mass] / [time] ** 2",
    lambda ureg, x, **kwargs: x / ureg.speed_of_light)
# specrad per wavenumber → specrad per frequency
sp2.add_transformation(
    "[mass] / [time] ** 2",
    "[length] * [mass] / [time] ** 3",
    lambda ureg, x, **kwargs: x * ureg.speed_of_light)
#sp2.add_transformation(
#    "[mass] / ([length] * [time] ** 3)",
#    "[length] * [mass] / [time] ** 3",
#    lambda ureg, x, **kwargs: )
def _R_to_bt(ureg, R, srf):
    """For use by pint, do not call directly, use q.to or SRF class."""
    if srf.lookup_table is None:
        srf.make_lookup_table()
    return ureg.Quantity(srf.L_to_T(R), 'K')
def _bt_to_R(ureg, T, srf):
    return ureg.Quantity(srf.blackbody_radiance(T, spectral=True))
_bt_to_R.__doc__ = _R_to_bt.__doc__
sp2.add_transformation(
    "[mass] / [time] ** 2",
    "[temperature]",
    _R_to_bt)
sp2.add_transformation(
    "[temperature]",
    "[mass] / [time] ** 2",
    _bt_to_R)
ureg.add_context(sp2)

radiance_units = {
    "si": ureg.W / (ureg.m**2 * ureg.sr * ureg.Hz),
    "ir": ureg.mW / (ureg.m**2 * ureg.sr * ureg.cm**-1)}

# add wavenumber/kayser to spectroscopy contexts for use with pint<0.8
ureg._contexts["spectroscopy"].add_transformation("[length]", "1/[length]",
    lambda ureg, x, **kwargs: 1/x)
ureg._contexts["spectroscopy"].add_transformation("1/[length]", "[length]",
    lambda ureg, x, **kwargs: 1/x)



"""ARTS C API Interface

This module provides a foreign function interface for the ARTS C API.
It defines the C structs used by the interface as ctypes.Structure
child classes as well as the return argument and return types of the
function provided by the C API.

Requirements
------------

The ARTS C API is provided by the arts_api.so library and is required by
the module. The module will check if the ``ARTS_BUILD_PATH`` variable is set
and assume the library can be found in the src subdirectory. If opening
the library fails loading the module will fail with an EnvironmentError.

Attributes:

    arts_api(CDLL): The ctypes library handle holding the ARTS C API.

"""

import ctypes as c
import logging
import os

import numpy as np
import scipy as sp

from typhon.environment import environ


logger = logging.getLogger(__name__)

################################################################################
# Version Requirements
################################################################################

arts_minimum_major    = 2
arts_minimum_minor    = 3
arts_minimum_revision = 1167

################################################################################
# Load ARTS C API
################################################################################

if environ.get("ARTS_BUILD_PATH") is None:
    raise EnvironmentError("ARTS_BUILD_PATH environment variable required to"
                           + " locate ARTS API.")

try:
    lib_path = os.path.join(environ.get("ARTS_BUILD_PATH"), "src",
                            "libarts_api.so")
    logger.info("Loading ARTS API from: " + lib_path)
    arts_api = c.cdll.LoadLibrary(lib_path)
except:
    raise EnvironmentError("Could not find ARTS API in your ARTS build path. "
                           "Did you install it?" + os.linesep +
                           "Typhon requires at least ARTS version "
                           f"{arts_minimum_major}.{arts_minimum_minor}."
                           f"{arts_minimum_revision}")

################################################################################
# Version Check
################################################################################

class VersionStruct(c.Structure):
    """
    The ARTS version is represented by 3 values of type long: the major, minor
    and revision number.
    """
    _fields_ = [("major", c.c_long),
                ("minor", c.c_long),
                ("revision", c.c_long)]

arts_api.get_version.argtypes = None
arts_api.get_version.restype  = VersionStruct

version = arts_api.get_version()
if (version.major, version.minor, version.revision) \
   < (arts_minimum_major, arts_minimum_minor, arts_minimum_revision):

    raise EnvironmentError("This typhon version requires at least arts-"
                           + str(arts_minimum_major) + "."
                           + str(arts_minimum_minor) + "."
                           + str(arts_minimum_revision) + " of ARTS.")

################################################################################
# Initialize API
################################################################################

arts_api.initialize()

################################################################################
# ARTS runtime environment manipulation
################################################################################

def find_controlfile(name):
    """ Recursively search arts include path for given file.
    Args:
        name(str): Name of the file.
    Raises:
        Exception: If the file cannot be found.
    Returns:
        path(str): The full path of the file.
    """
    paths = arts_include_path + [os.getcwd()]
    path  = None

    for p in paths:
        if os.path.isfile(os.path.join(p, name)):
            path = os.path.join(p, name)
    if (path):
        return path
    else:
        raise Exception("File " + name + " not found. Search path was:\n "
                        + str(paths))

def include_path_push(path):
    """
    Add path to include path of the ARTS runtime.

    Args:
        path(str): Path to add to the ARTS include path.
    """
    arts_api.include_path_push(c.c_char_p(path.encode()))

def include_path_pop():
    """
    Remove most recently added include path.
    """
    arts_api.include_path_pop()

def data_path_push(path):
    """
    Add path to data path of the ARTS runtime.

    Args:
        path(str): Path to add to the ARTS data path.
    """
    arts_api.data_path_push(c.c_char_p(path.encode()))

def data_path_pop():
    """
    Remove most recently added data path.
    """
    arts_api.data_path_pop()


################################################################################
# ctypes Structures
################################################################################

class VariableStruct(c.Structure):
    """
    A (symbolic) ARTS workspace variable is represented using a struct containing
    pointers to the name and description of the method as well as the group id,
    i.e. the Index variable encoding the type of the variable.
    """
    _fields_ = [("name", c.c_char_p),
                ("description", c.c_char_p),
                ("group", c.c_long)]


class VariableValueStruct(c.Structure):
    """
    The ARTS C API uses C-structs to transfer workspace variables from and to the
    ARTS runtime. The VariableValueStruct class is used to access and create these
    structs in Python code. The fields of the C-struct can be accessed directly as
    attributes of the VariableValueStruct object.
    """
    _fields_ = [("ptr", c.c_void_p),
                ("initialized", c.c_bool),
                ("dimensions", 7 * c.c_long),
                (("inner_ptr"), c.POINTER(c.c_int)),
                (("outer_ptr"), c.POINTER(c.c_int))]

    @classmethod
    def empty(cls):
        s = VariableValueStruct(0)
        s.ptr = 0
        return s

    def __init__(self, value):
        """ Create a VariableValue struct from a python object.

        This functions creates a variable value struct from a python object so
        that it can be passed to the C API. If the type of the object is not
        supported, the data pointer will be NONE.

        The built-in Python types that are currently supported are:

            - int
            - float
            - string
            - numpy.ndarray
            - lists of int and lists of string

        User defined classes are supported through a generic interface. The constructor
        looks for an attribute function __to_value_struct__, which should return a dictionary
        containing the value associated with the fields of the C-struct.

        Args:
            value(object): The python object to represent as a VariableValue struct.

        """
        ptr = None
        initialized = True
        dimensions  = [0] * 7

        self._temporaries = []

        # Generic interface
        if hasattr(value, "__to_value_struct__"):
            d = value.__to_value_struct__()
            if "ptr" in d:
                ptr = d["ptr"]
            if "dimensions" in d:
                dimensions = d["dimensions"]
        # Index
        elif isinstance(value, np.long):
            ptr = c.cast(c.pointer(c.c_long(value)), c.c_void_p)
        # Numeric
        elif isinstance(value, (float, np.double)):
            temp = np.float64(value)
            ptr = c.cast(c.pointer(c.c_double(temp)), c.c_void_p)
        # String
        elif isinstance(value, str):
            ptr = c.cast(c.c_char_p(value.encode()), c.c_void_p)
        # Vector, Matrix
        elif isinstance(value, np.ndarray):
            # arrays need to be contiguous when passed to the ARTS API
            value = np.ascontiguousarray(value)

            if value.dtype == np.float64:
                ptr = value.ctypes.data
                for i in range(value.ndim):
                    dimensions[i] = value.shape[i]
        # Scipy sparse matrices
        elif sp.sparse.issparse(value):
            m = sp.sparse.coo_matrix(value)
            self._temporaries += [m]
            dimensions[0] = m.shape[0]
            dimensions[1] = m.shape[1]
            dimensions[2] = m.nnz
            ptr = m.data.ctypes.data
            self.inner_ptr = c.cast(m.row.ctypes.data, c.POINTER(c.c_int))
            self.outer_ptr = c.cast(m.col.ctypes.data, c.POINTER(c.c_int))
        # Array of String or Integer
        elif isinstance(value, list):
            if not value:
                raise ValueError("Empty lists currently not supported.")
            ps = []
            if isinstance(value[0], str):
                for s in value:
                    ps.append(c.cast(c.c_char_p(s.encode()), c.c_void_p))
                p_array = (c.c_void_p * len(value))(*ps)
                ptr = c.cast(c.pointer(p_array), c.c_void_p)
            if isinstance(value[0], np.long):
                ptr = c.cast(c.pointer((c.c_long * len(value))(*value)), c.c_void_p)
            dimensions[0] = len(value)

        self._value = value
        self.ptr = ptr
        self.initialized = initialized
        self.dimensions  = (c.c_long * 7)(*dimensions)

class CovarianceMatrixBlockStruct(c.Structure):
    """
    c struct representing block of covariance matrices.

    The indices field holds the indices of the corresponding
    retrieval quantities.

    The position field holds the row and column indices of the
    left- and upper-most element of the block w.r.t. the full
    covariance matrix.

    The dimension field hold the number of rows and columns of
    the block.

    The ptr field hold the pointer to the dense matrix data or
    to the element pointer of the sparse matrix that represents
    the block.

    The inner and outer pointer fields are null if the block is
    represented by a dense matrix. Otherwise these contain the
    pointers to the index array of the sparse matrix of which the
    block consists.
    """
    _fields_ = [("indices", 2 * c.c_long),
                ("position", 2 * c.c_long),
                ("dimensions", 2 * c.c_long),
                ("ptr", c.c_void_p),
                ("nnz", c.c_long),
                ("inner_ptr", c.POINTER(c.c_int)),
                ("outer_ptr", c.POINTER(c.c_int))]

class MethodStruct(c.Structure):
    """
    The method struct holds the internal index of the method (id), pointers
    to the null-terminated strings holding name and description, the number
    of generic inputs (n_g_in) and a pointer to the array holding the group ids
    of the output types, as well as the number of generic outputs and their types.
    """
    _fields_ = [("id", c.c_ulong),
                ("name", c.c_char_p),
                ("description", c.c_char_p),
                # Output
                ("n_out", c.c_ulong),
                ("outs", c.POINTER(c.c_long)),
                # Generic Output
                ("n_g_out", c.c_ulong),
                ("g_out_types", c.POINTER(c.c_long)),
                # Input
                ("n_in", c.c_ulong),
                ("ins", c.POINTER(c.c_long)),
                # Generic Input
                ("n_g_in", c.c_ulong),
                ("g_in_types", c.POINTER(c.c_long))]

# TODO: Check if can be used as constructor
def variable_value_factory(value):
    """ Create a VariableValue struct from a python object.

    This functions creates a variable value struct from a python object so that it
    can be passed to the C API. If the type of the object is not supported, the data
    pointer will be NULL.

    Args:
        value(object): The python object to represent as a VariableValue struct.

    TODO: Add proper error handling.
    """
    return VariableValueStruct(value)

################################################################################
# Function Arguments and Return Types
################################################################################

# Create ArtsWorkspace and return handle.
arts_api.create_workspace.argtypes = [c.c_long, c.c_long]
arts_api.create_workspace.restype  = c.c_void_p

# Destroy ArtsWorkspace instance from handle.
arts_api.destroy_workspace.argtypes = [c.c_void_p]
arts_api.destroy_workspace.restype   = None

# Include path manipulation.
arts_api.include_path_push.restype = None
arts_api.include_path_push.argtypes = [c.c_char_p]

# Data path manipulation.
arts_api.include_path_pop.restype = None
arts_api.include_path_pop.argtypes = None

arts_api.data_path_push.restype = None
arts_api.data_path_push.argtypes = [c.c_char_p]

arts_api.data_path_pop.restype = None
arts_api.data_path_pop.argtypes = None

# Set include ad data path of the arts runtime.
arts_api.get_error.restype  = c.c_char_p
arts_api.get_error.argtypes = None

# Agendas
#
#
arts_api.create_agenda.argtypes = [c.c_char_p]
arts_api.create_agenda.restype  = c.c_void_p

arts_api.agenda_add_method.argtypes = [c.c_void_p, c.c_long,
                                       c.c_ulong, c.POINTER(c.c_long),
                                       c.c_ulong,c.POINTER(c.c_long)]
arts_api.agenda_add_method.restype  = None

arts_api.agenda_clear.argtypes = [c.c_void_p]
arts_api.agenda_clear.restype  = None

arts_api.agenda_insert_set.argtypes = [c.c_void_p, c.c_void_p, c.c_long, c.c_long]
arts_api.agenda_insert_set.restype = None

arts_api.agenda_append.argtypes = [c.c_void_p, c.c_void_p]
arts_api.agenda_append.restype = None

arts_api.parse_agenda.argtypes = [c.c_char_p]
arts_api.parse_agenda.restype  = c.c_void_p

arts_api.execute_agenda.argtypes = [c.c_void_p, c.c_void_p]
arts_api.execute_agenda.restype  = c.c_char_p

arts_api.destroy_agenda.argtypes = [c.c_void_p]
arts_api.destroy_agenda.restype  = None

# Groups
#
# Returns the number of WSV groups.
arts_api.get_number_of_groups.argtypes = None
arts_api.get_number_of_groups.restype  = c.c_ulong

# Return pointer to the name of the group with given index.
arts_api.get_group_name.argtypes = [c.c_long]
arts_api.get_group_name.restype  = c.c_char_p

# Variables
#
# Returns the number of (symbolic) workspace variable.
arts_api.get_number_of_variables.restype  = c.c_ulong
arts_api.get_number_of_variables.argtypes = None

# Returns workspace variable with index c_long as VariableStruct.
arts_api.lookup_workspace_variable.argtypes = [c.c_char_p]
arts_api.lookup_workspace_variable.restype  = c.c_long

# Returns workspace variable with index c_long as VariableStruct.
arts_api.get_variable.argtypes = [c.c_long]
arts_api.get_variable.restype  = VariableStruct

# Return pointer to variable value in a given workspace in the form of a VariableValueStruct.
arts_api.get_variable_value.argtypes = [c.c_void_p, c.c_long, c.c_long]
arts_api.get_variable_value.restype  = VariableValueStruct

# Set variable value in workspace given a workspace handle, the variable id, the group id
# and a VariableValueStruct
arts_api.set_variable_value.argtypes = [c.c_void_p, c.c_long, c.c_long, VariableValueStruct]
arts_api.set_variable_value.restype  =  c.c_char_p

# Adds a value of a given group to a given workspace.
arts_api.add_variable.restype  = c.c_long
arts_api.add_variable.argtypes = [c.c_void_p, c.c_long, c.c_char_p]

# Remove given variable from workspace.
arts_api.erase_variable.restype  = None
arts_api.erase_variable.argtypes = [c.c_void_p, c.c_long, c.c_long]

# Methods
#
# Returns the number of (symbolic) workspace variable.
arts_api.get_number_of_methods.restype  = c.c_ulong
arts_api.get_number_of_methods.argtypes = None

# Returns workspace variable with index c_long as VariableStruct.
arts_api.get_method.argtypes = [c.c_long]
arts_api.get_method.restype  = MethodStruct

# Return Pointer to name of jth generic output parameter of a given WSM.
arts_api.get_method_g_out.argtypes = [c.c_long, c.c_long]
arts_api.get_method_g_out.restype  = c.c_char_p

# Return Pointer to name of jth generic input parameter of a given WSM.
arts_api.get_method_g_in.argtypes = [c.c_long, c.c_long]
arts_api.get_method_g_in.restype  = c.c_char_p

# Return pointer to the default value of the jth generic input of a given WSM.
arts_api.get_method_g_in_default.argtypes = [c.c_long, c.c_long]
arts_api.get_method_g_in_default.restype  = c.c_char_p

# Return block from covariance matrix.
arts_api.get_covariance_matrix_block.argtypes = [c.c_void_p, c.c_long, c.c_bool]
arts_api.get_covariance_matrix_block.restype = CovarianceMatrixBlockStruct

# Execute a given workspace method.
arts_api.execute_workspace_method.argtypes = [c.c_void_p,
                                              c.c_long,
                                              c.c_ulong,
                                              c.POINTER(c.c_long),
                                              c.c_ulong,
                                              c.POINTER(c.c_long)]
arts_api.execute_workspace_method.restype  = c.c_char_p

# Print method documentation.
arts_api.method_print_doc.argtypes = [c.c_long]
arts_api.method_print_doc.restype  = c.c_char_p

# Callback insertion
arts_api.callbacks = []
arts_api.agenda_insert_callback.argtypes = [c.c_void_p, c.c_void_p]
arts_api.agenda_insert_callback.restype  = None

################################################################################
# Setup ARTS Environment
################################################################################

try:
    arts_include_path = environ.get("ARTS_INCLUDE_PATH").split(":")
except:
    arts_include_path = []

try:
    arts_data_path = environ.get("ARTS_DATA_PATH").split(":")
except:
    arts_data_path = []

# Set runtime parameters
for p in arts_include_path:
    include_path_push(p)

for p in arts_data_path:
    data_path_push(p)

include_path_push(os.getcwd())
data_path_push(os.getcwd())



# -*- coding: utf-8 -*-
"""Testing the environment/configuration handler.
"""
import os
from copy import copy

import pytest

from typhon import environ
from typhon.config import conf


class TestEnvironment:
    """Testing the environment handler."""
    def setup_method(self):
        """Run all test methods with empty config and environment."""
        self.env = copy(os.environ)
        self.conf = copy(conf)

        os.environ = {}
        conf.clear()

    def teardown_method(self):
        """Restore config and environment."""
        os.environ = self.env
        conf = self.conf

    def test_get_environment_variables(self):
        """Test if environment variables are considered."""
        os.environ['TYPHON_ENV_TEST'] = 'TEST_VALUE'

        assert environ['TYPHON_ENV_TEST'] == 'TEST_VALUE'

    def test_set_environment_variables(self):
        """Test if environment variables are updated."""
        environ['TYPHON_ENV_TEST'] = 'TEST_VALUE'

        assert os.environ['TYPHON_ENV_TEST'] == 'TEST_VALUE'

    def test_undefined_variable(self):
        """Test behavior for undefined variables."""
        with pytest.raises(KeyError):
            dummy = environ['TYPHON_ENV_TEST']

    def test_membership(self):
        """Test the membership check."""
        os.environ['TYPHON_ENV_TEST'] = 'TEST_VALUE'

        assert 'TYPHON_ENV_TEST' in environ

    def test_membership_negative(self):
        """Test the membership check (negative)."""
        assert 'TYPHON_ENV_TEST' not in environ



# -*- coding: utf-8 -*-

"""Functions to create plots using matplotlib.
"""

import numpy as np
import matplotlib.pyplot as plt


__all__ = [
    'plot_ppath',
    'plot_ppath_field',
    'ppath_field_minmax_posbox',
    'adjust_ppath_field_box_by_minmax',
    'plot_ppath_field_zenith_coverage_per_gp_p',
]


def plot_ppath(ppath, planetary_radius=1, lat_is_x=True, scale_alt=1000, ax=None):
    """ Return the ARTS Ppath plotted on the surface

    Parameters:
        ppath:
            (Ppath or Workspace)  Propagation path or Wosrkspace with the path
        planetary_radius:
            (Numeric)  Spherical radius of the planetary body (if ppath_field is not Workspace)
        lat_is_x:
            (Boolean) Have lat or lon act as the x-coordinate?
        scale_alt:
            (Numeric) Divide altitude and planetary_radius by this value
        ax:
            (AxesSubplot) axes to plot into.  Assumes polar projection.  If None,
            creates a plt.figure(projection='polar') and draws on its axis

        Example:
            >>> arts = typhon.arts.workspace.Workspace()
            >>> ###################################################
            >>> # Run ARTS simulation that sets an external ppath #
            >>> ###################################################
            >>> plot_ppath(arts)
    """

    try:
        planetary_radius = ppath.refellipsoid.value[0]
    except:
        pass

    try:
        ppath = ppath.ppath.value
    except:
        pass

    alt, lat, lon, za, aa = ppath.alt_lat_lon_za_aa()

    if lat_is_x:
        x = lat
    else:
        x = lon

    if ax is None:
        ax = plt.subplot(111, projection='polar')

    ax.plot(np.deg2rad(x), alt/scale_alt)
    ax.set_rorigin(-planetary_radius/scale_alt)

    return ax


def plot_ppath_field(ppath_field, planetary_radius=1, lat_is_x=True,
                     scale_alt=1000, subplots=1, auto_adjust=True, axes=None):
    """ Return the ARTS Ppath plotted on the surface

    Parameters:
        ppath_field:
            (Ppath or Workspace)  Propagation path or Wosrkspace with the path
        planetary_radius:
            (Numeric)  Spherical radius of the planetary body (if ppath_field is not Workspace)
        lat_is_x:
            (Boolean) Have lat or lon act as the x-coordinate?
        scale_alt:
            (Numeric) Divide altitude and planetary_radius by this value
        subplots:
            (Index)  Divides the ppath_field into this many equally long
            segments, so that the first len(ppath_field) // subplots plots
            are drawn on the same surface and so forth
        auto_adjust:
            (Boolean)  If true calls adjust_ppath_field_box_by_minmax
        axes:
            Either list of subplots, or None.

    Returns:
        list of axis of the length of subplots
    """

    try:
        planetary_radius = ppath_field.refellipsoid.value[0]
    except:
        pass

    try:
        ppath_field = ppath_field.ppath_field.value
    except:
        pass

    N = len(ppath_field)
    if N % subplots != 0:
        raise RuntimeError("Bad inputs. Only for evenly divisible numbers")

    # Number of drawings per subplot
    n = N // subplots

    n1 = int(np.ceil(np.sqrt(subplots)))
    n2 = n1 if np.isclose(n1*n1, subplots) or n1*n1 > subplots else n1 + 1

    if axes is None:
        fig = plt.gcf()
        axes = []
        for i in range(subplots):
            axes.append(fig.add_subplot(n1, n2, i+1, projection='polar'))

    assert subplots == len(axes), "Must have same number of axes as subplots"

    for i in range(subplots):
        for j in range(n):
            plot_ppath(ppath_field[j+n*i], planetary_radius, lat_is_x,
                       scale_alt, ax=axes[i])

    if auto_adjust:
        adjust_ppath_field_box_by_minmax(axes, ppath_field, lat_is_x, scale_alt, 0.05)

    return axes


def ppath_field_minmax_posbox(ppath_field):
    """ Return the minimum and maximum of all pos variables of ppath_field

    Parameters:
        ppath_field:
            (Ppath or Workspace)  Propagation path or Wosrkspace with the path

    Returns:
        min(alt), max(alt), min(lat), max(lat), min(lon), max(lon)
    """

    try:
        ppath_field = ppath_field.ppath_field.value
    except:
        pass

    minalt = minlat = minlon = np.infty
    maxalt = maxlat = maxlon = -np.infty

    for ppath in ppath_field:
        alt, lat, lon, za, aa = ppath.alt_lat_lon_za_aa()

        minalt = min(alt.min(), minalt)
        maxalt = max(alt.max(), maxalt)

        minlat = min(lat.min(), minlat)
        maxlat = max(lat.max(), maxlat)

        minlon = min(lon.min(), minlon)
        maxlon = max(lon.max(), maxlon)

    return minalt, maxalt, minlat, maxlat, minlon, maxlon


def adjust_ppath_field_box_by_minmax(axes, ppath_field, lat_is_x=True,
                                     scale_alt=1000, extrap=0.05):
    """ Adjusts the axis of plotted ppath_field to

    Parameters:
        axes:
            Plotted axes of ppath_field
        ppath_field:
            (Ppath or Workspace)  Propagation path or Wosrkspace with the path
        planetary_radius:
            (Numeric)  Spherical radius of the planetary body (if ppath_field is not Workspace)
        lat_is_x:
            (Boolean) Have lat or lon act as the x-coordinate?
        scale_alt:
            (Numeric) Divide altitude and planetary_radius by this value
        extrap:
            (Numeric) How large an extrapolation of the box in terms of full
            posbox minmax
    """

    minalt, maxalt, minlat, maxlat, minlon, maxlon = ppath_field_minmax_posbox(ppath_field)

    if lat_is_x:
        xmin = minlat
        xmax = maxlat
        max_diff = 180
    else:
        xmin = minlon
        xmax = maxlon
        max_diff = 360
    dx = (xmax-xmin) * extrap

    xmin -= dx
    xmax += dx

    if (xmax-xmin) > max_diff:
        return  # no adjustment possible

    for ax in axes:
        ax.set_thetamin(xmin)
        ax.set_thetamax(xmax)


def wzeniths(zeniths):
    n = len(zeniths)
    if not n:
        return np.array([0, 180]), np.array([1, 1])
    inds = np.argsort(zeniths)
    zaz = np.deg2rad(zeniths[inds])
    cz = np.cos(zaz)

    wz = np.zeros((2*n))
    za = np.zeros((2*n))
    for i in range(n-1):
        N = i*2
        za[N:N+2] = zaz[i:i+2]
        wz[0+N] = wz[1+N] = 0.5 * (cz[i] - cz[i+1])

    return za, wz


def plot_ppath_field_zenith_coverage_per_gp_p(ppath_field, scale_alt=1000, axes=None):
    """Plots the zenith angle coverage of a ppath_field for all the altitudes
    in the field.

    Parameters:
        ppath_field:
            (Ppath or Workspace)  Propagation path or Wosrkspace with the path
        axes:
            Either list of subplots, or None.

    Returns:
        list of axis, list of altitudes, list of maximum weights
    """
    try:
        ppath_field = ppath_field.ppath_field.value
    except:
        pass

    alt = []
    gps = []
    for path in ppath_field:
        for i in range(path.np):
            if path.gp_p[i] not in gps:
                gps.append(path.gp_p[i])
                alt.append(path.pos[i, 0])
    alt = np.array(alt)

    # Size of subplots
    N = len(alt)

    zas = np.full((N), None)
    for path in ppath_field:
        for i in range(path.np):
            ip = np.where(path.pos[i, 0] == alt)[0][0]
            if zas[ip]:
                zas[ip].append(path.los[i, 0])
            else:
                zas[ip] = [path.los[i, 0]]

    # Sub-plot grid
    n1 = int(np.ceil(np.sqrt(N)))
    n2 = n1 if np.isclose(n1*n1, N) or n1*n1 > N else n1 + 1

    if axes is None:
        fig = plt.gcf()
        axes = []
        for i in range(N):
            axes.append(fig.add_subplot(n1, n2, i+1, projection='polar'))

    assert N == len(axes), "Must have same number of axes as subplots"

    inds = np.argsort(alt)

    wei = []
    alts = []
    for i in range(N):
        pos = inds[i]
        za, wz = wzeniths(np.array(zas[pos]))

        wei.append(wz.max())
        alts.append(alt[pos])
        axes[i].plot(za, wz,'k')
        axes[i].set_rorigin(-0.25/np.pi)
        axes[i].set_thetamin(0)
        axes[i].set_thetamax(180)
        axes[i].set_rgrids([0, wz.max()*1.05], ['', ''])
        axes[i].set_thetagrids([0, 90, 180], ['Zenith', 'Limb', 'Nadir'])
        axes[i].set_title("Alt {} km".format(alt[pos]/scale_alt))

    return axes, alts, wei



# -*- coding: utf-8 -*-
"""Testing the functions in typhon.plots.
"""
import os

import pytest

from typhon import plots


class TestPlots:
    """Testing the plot functions."""
    def test_figsize(self):
        """Test golden ratio for figures sizes."""
        ret = plots.figsize(10)

        assert ret == (10, 6.1803398874989481)

    def test_get_subplot_arrangement(self):
        """Test the determination of subplot arrangements."""
        shape = plots.get_subplot_arrangement(8)
        assert shape == (3, 3)

    def test_get_style_path_method(self):
        assert os.path.isfile(plots.styles.get('typhon'))

    def test_get_style_path_call(self):
        assert os.path.isfile(plots.styles('typhon'))

    def test_get_style_path_default(self):
        assert os.path.isfile(plots.styles())

    def test_undefined_style(self):
        with pytest.raises(ValueError):
            plots.styles.get('Undefined Stylesheet Name')

    def test_available_styles(self):
        style_paths = plots.styles.available

        assert isinstance(style_paths, list)
        assert len(style_paths) > 0
        assert all(os.path.isfile(plots.styles(s)) for s in style_paths)



# -*- coding: utf-8 -*-
"""Custom tick formatter. """
import matplotlib.pyplot as plt
from matplotlib.ticker import LogFormatter, FuncFormatter

from typhon import constants


__all__ = [
    'set_xaxis_formatter',
    'set_yaxis_formatter',
    'HectoPascalFormatter',
    'HectoPascalLogFormatter',
    'ScalingFormatter',
]


def set_xaxis_formatter(formatter, ax=None):
    """Set given formatter for major and minor xticks."""
    if ax is None:
        ax = plt.gca()

    ax.xaxis.set_major_formatter(formatter)
    ax.xaxis.set_minor_formatter(formatter)


def set_yaxis_formatter(formatter, ax=None):
    """Set given formatter for major and minor yticks."""
    if ax is None:
        ax = plt.gca()

    ax.yaxis.set_major_formatter(formatter)
    ax.yaxis.set_minor_formatter(formatter)


def ScalingFormatter(scaling=1, fmtstr='{x:g}'):
    """Provide a ticklabel formatter that applies scaling.

    Parameters:
        scaling (float or string): Scaling that is applied to all labels.
            If `str`, try to find corresponding scale in `typhon.constants`.
        fmtstr (str): Format string used to create label text.
            The field `x` is replaced with the scaled label value.

    Returns:
        matplotlib.ticker.FuncFormatter: Formatter.

    Examples:

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        from typhon.plots import (set_yaxis_formatter, ScalingFormatter)


        y = 1e6 * np.random.randn(100)

        fig, ax = plt.subplots()
        ax.plot(y)
        ax.set_ylabel('y')
        ax.set_title('default')

        fig, ax = plt.subplots()
        ax.plot(y)
        set_yaxis_formatter(ScalingFormatter(scaling=1e6))
        ax.set_ylabel('y in millions')
        ax.set_title('float scaling')

        fig, ax = plt.subplots()
        ax.plot(y)
        set_yaxis_formatter(ScalingFormatter(scaling=1e6, fmtstr='{x:g}M'))
        ax.set_ylabel('y')
        ax.set_title('float scaling and custom label')

        fig, ax = plt.subplots()
        ax.plot(y)
        set_yaxis_formatter(ScalingFormatter(scaling='kilo', fmtstr='{x:g}k'))
        ax.set_ylabel('y')
        ax.set_title('string scaling and custom label')

        plt.show()

    """
    # Try to find string scaling as attributes in `typhon.constants`.
    if isinstance(scaling, str):
        scaling = getattr(constants, scaling)

    @FuncFormatter
    def formatter(x, pos):
        return fmtstr.format(x=x / scaling)

    return formatter


def HectoPascalFormatter():
    """Creates hectopascal labels for pascal input.

    Note:
        Simple wrapper for :func:`~typhon.plots.ScalingFormatter`.

    Examples:

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import typhon
        from typhon.plots import (set_yaxis_formatter, HectoPascalFormatter)


        p = typhon.math.nlogspace(1000e2, 0.1e2, 50)

        fig, ax = plt.subplots()
        ax.plot(np.exp(p / p[0]), p)
        ax.invert_yaxis()
        set_yaxis_formatter(HectoPascalFormatter())

        plt.show()

    See also:
            :func:`~typhon.plots.HectoPascalLogFormatter`
                Creates logarithmic hectopascal labels for pascal input.
    """
    return ScalingFormatter('hecto')


class HectoPascalLogFormatter(LogFormatter):
    """Creates logarithmic hectopascal labels for pascal input.

    This class can be used to create axis labels on the hectopascal scale for
    values plotted in pascals. It is handy in combination with plotting in
    logscale.

    Examples:

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        import typhon
        from typhon.plots import (set_yaxis_formatter, HectoPascalLogFormatter)


        p = typhon.math.nlogspace(1000e2, 0.1e2, 50)  # pressue in Pa

        fig, ax = plt.subplots()
        ax.semilogy(np.exp(p / p[0]), p)
        ax.invert_yaxis()
        set_yaxis_formatter(typhon.plots.HectoPascalLogFormatter())

        plt.show()

    See also:
            :func:`~typhon.plots.HectoPascalFormatter`
                Creates hectopascal labels for pascal input.
    """
    # TODO (lkluft): Hack to preserve automatic toggling to minor ticks for
    # log-scales introduced in matplotlib 2.0.
    # Could be replaced with a decent Formatter subclass in the future.
    def _num_to_string(self, x, vmin, vmax):
        return '{:g}'.format(x / 100)





# -*- coding: utf-8 -*-

"""Functions directly related to spectroscopy.
"""
import numpy as np
from scipy import interpolate

import typhon.constants as constants


__all__ = [
    'linewidth',
    'doppler_broadening',
    'boltzmann_level',
    'stimulated_emission',
]


def linewidth(f, a):
    """Calculate the full-width at half maximum (FWHM) of an absorption line.

    Parameters:
        f (ndarray): Frequency grid.
        a (ndarray): Line properties
            (e.g. absorption coefficients or cross-sections).

    Returns:
        float: Linewidth.

    Examples:
        >>> f = np.linspace(0, np.pi, 100)
        >>> a = np.sin(f)**2
        >>> linewidth(f, a)
        1.571048056449009
    """
    s = interpolate.UnivariateSpline(f, a - np.max(a)/2, s=0)
    return float(np.diff(s.roots()))


def doppler_broadening(t, f0, m):
    """Calculate the doppler broadening half-width half-maximum

    .. math::
        \\gamma_D(T) = \\sqrt{ \\frac{2\\log(2) k_B T}{mc^2} } f_0

    Parameters:
        t (float or ndarray): Temperature [Kelvin]

        f0 (float or like temperature): Central frequency [Hertz/invcm]

        m (float or like temperature): Mass [kilogram]

    Returns
        hwhm (like temperature): Half-width half-maximum [Hertz/invcm]
    """

    return np.sqrt(2 * constants.boltzmann * t * np.log(2) /
                   (m * constants.speed_of_light**2)) * f0


def boltzmann_level(elow, t, t0):
    """Computes the Boltzmann level function

    .. math::
        K_1 = \\exp\\left(\\frac{E_l \\left[T-T_0\\right]}{k_B T T_0}\\right),

    where :math:`k_B` is the Boltzmann constant.

    All ndarrays must be of same size, any of the inputs can be ndarray

    Parameters:
        elow (float or ndarray): Lower state energy level [J]

        t (float or ndarray): Temperature [Kelvin]

        t0 (float or ndarray): Line temperature [Kelvin]

    Returns
        K1 (like input): How much Boltzmann statistics feeds the transition

    .. math::
        S(T) = S(T_0)K_1K_2 \\frac{Q(T_0)}{Q(T)}
    """
    return np.exp(elow * (t - t0) / (constants.boltzmann * t * t0))


def stimulated_emission(f0, t, t0):
    """Computes the stimulated emission function

    .. math::
        K_2 = \\frac{1 - \\exp\\left( - \\frac{h f_0}{k_B T}\\right)}
        {1 - \\exp\\left( - \\frac{h f_0}{k_B T_0}\\right)},

    with Planck constant :math:`h` and Boltzmann constant :math:`k_B`.

    Parameters:
        f0 (float or ndarray): Line frequency [Hz]

        t (float or ndarray): Temperature [Kelvin]

        t0 (float or ndarray): Line temperature [Kelvin]

    Returns
        K2 (like input): How stimulated the emission is

    .. math::
        S(T) = S(T_0)K_1K_2 \\frac{Q(T_0)}{Q(T)}
    """
    return (1. - np.exp(- constants.planck * f0/(constants.boltzmann * t))) / \
        (1. - np.exp(- constants.planck * f0/(constants.boltzmann * t0)))



"""Miscellaneous unit-aware tools
"""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

from .common import ureg

import operator
import numpy
import xarray
class UnitsAwareDataArray(xarray.DataArray):
    """Like xarray.DataArray, but transfers units
    """

    # need to keep both __array_wrap__ and __array_ufunc__.  Although the
    # former supersedes the latter, xarrays methods explicitly call the
    # former sometimes.
    def __array_wrap__(self, obj, context=None):
        new_var = super().__array_wrap__(obj, context)
        if self.attrs.get("units"):
            if context:
                # some ufuncs like exp return regular numpy floats rather
                # than (dimensionless) quantities, see
                # https://github.com/hgrecco/pint/issues/482
                q = context[0](ureg.Quantity(1, self.attrs.get("units")))
                try:
                    u = q.u
                except AttributeError:
                    if (ureg(self.attrs["units"]).dimensionless or
                        new_var.dtype.kind == "b"):
                        # expected, see # https://github.com/hgrecco/pint/issues/482
                        u = ureg.dimensionless
                    else:
                        raise
                    # for exp and log, values are not set correctly.  I'm
                    # not sure why.  Perhaps related to
                    # https://github.com/hgrecco/pint/issues/493
                    new_var.values = context[0](ureg.Quantity(self.values, self.units))
                new_var.attrs["units"] = str(u)
            else: # unary operators always retain units?
                new_var.attrs["units"] = str(self.attrs.get("units"))
        return new_var

    def _apply_binary_op_to_units(self, func, other, x):
        if self.attrs.get("units"):
            x.attrs["units"] = str(func(ureg.Quantity(1, self.attrs["units"]),
                                    ureg.Quantity(1, getattr(other, "units", "1"))).u)
        return x

    def _apply_rbinary_op_to_units(self, func, other, x):
        if self.attrs.get("units"):
            x.attrs["units"] = str(func(ureg.Quantity(1, getattr(other, "units", "1")),
                                    ureg.Quantity(1, self.attrs["units"]),).u)
        return x

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        new_var = super().__array_ufunc__(ufunc, method, *args, **kwargs)
        # make sure we're still UADA
        new_var = self.__class__(new_var)
        if self.attrs.get("units"):
            if method == "__call__":
                q = ufunc(ureg.Quantity(1, self.attrs.get("units")))
                try:
                    u = q.u
                except AttributeError:
                    if (ureg(self.attrs["units"]).dimensionless or
                        new_var.dtype.kind == "b"):
                        # expected, see https://github.com/hgrecco/pint/issues/482
                        u = ureg.dimensionless
                    else:
                        raise
                    # for exp and log, values are not set correctly.  I'm
                    # not sure why.  Perhaps related to
                    # https://github.com/hgrecco/pint/issues/493
                    new_var.values = ufunc(ureg.Quantity(self.values, self.units))
                new_var.attrs["units"] = str(u)
            else: # unary operators? always retain units?
                raise NotImplementedError("Not implented")
                new_var.attrs["units"] = str(self.attrs.get("units"))
                    
        return new_var

    # pow is different because resulting unit depends on argument, not on
    # unit of argument
    def __pow__(self, other):
        x = super().__pow__(other)
        if self.attrs.get("units"):
            x.attrs["units"] = str(pow(ureg.Quantity(1, self.attrs["units"]),
                                   ureg.Quantity(other, getattr(other, "units", "1"))).u)
        return x

    def __rpow__(self, other):
        x = super().__rpow__(other)
        if self.attrs.get("units"):
            # NB: this should fail if we're not dimensionless
            x.attrs["units"] = str(pow(
                ureg.Quantity(other, getattr(other, "units", "1")),
                ureg.Quantity(1, self.attrs["units"])).u)
        return x

    def _reduce(self, attr, *args, **kwargs):
        # unit on diff, sum etc. is a bit involved as it should depend on the
        # coordinates of the dimension along which the reduction is taken, but
        # as it is all discrete I consider the unit to stay the same
        u = self.attrs.get("units")
        d = getattr(super(), attr)(*args, **kwargs)
        if u is not None:
            d.attrs["units"] = u
        return d

    # not sure if/how I can/should define those in a loop.  Should also have
    # std, min, max, mean, probably others.
    def diff(self, *args, **kwargs):
        return self._reduce("diff", *args, **kwargs)

    def sum(self, *args, **kwargs):
        return self._reduce("sum", *args, **kwargs)

    def mean(self, *args, **kwargs):
        return self._reduce("mean", *args, **kwargs)

    def median(self, *args, **kwargs):
        return self._reduce("median", *args, **kwargs)

    def to(self, new_unit, *contexts, **kwargs):
        """Convert to other unit.

        See corresponding method in pint.
        """
        x = self.copy()
        x.values = ureg.Quantity(self.values, self.attrs["units"]).to(
            new_unit, *contexts, **kwargs).m
        x.attrs["units"] = str(new_unit)
        return x

    def to_root_units(self):
        """Equivalent of pint's Quantity.to_root_units
        """
        x = self.copy()
        q = ureg.Quantity(self.values, self.attrs["units"]).to_root_units()
        x.values = q.m
        x.attrs["units"] = str(q.u)
        return x

tp_all = ["add", "sub", "mul", "matmul", "truediv", "floordiv", "mod",
          "divmod"]
for tp in tp_all + ["r"+x for x in tp_all]:
    meth = "__{:s}__".format(tp)
    def func(self, other, meth=meth, tp=tp):
        rat = (ureg.Quantity(1, getattr(self, "units", "1")) /
               ureg.Quantity(1, getattr(other, "units", "1")))
        if (rat.u.dimensionless and rat.m != 1
#        if (ureg.Quantity(1, getattr(self, "units", "1")) /
#            ureg.Quantity(1, getattr(other, "units", "1")))
#        if (ureg.Quantity(1, getattr(self, "units", "1")).to_root_units()
#            == ureg.Quantity(1, getattr(other, "units", "1")).to_root_units()
#            and ureg.Quantity(1, self.units).m != ureg.Quantity(1, self.units).to("1").m
#            and ureg.Quantity(1, self.units).m != ureg.Quantity(1, self.units).to("1").m
#            and getattr(self, "units", "1") != getattr(other, "units", "1")
            and meth[2:-2] not in ("mul", "rmul", "floordiv", "truediv")):
            raise NotImplementedError("Do not add arrays with same units "
                "but different prefixes, currently buggy, see #150.")
        x = getattr(super(UnitsAwareDataArray, self), meth)(other)
        if tp.startswith("r"):
            return self._apply_rbinary_op_to_units(getattr(operator, tp[1:]), other, x)
        else:
            return self._apply_binary_op_to_units(getattr(operator, tp), other, x)
    func.__name__ = meth
    setattr(UnitsAwareDataArray, meth, func)
del func, tp_all



# -*- coding: utf-8 -*-

"""Implementation of SPARE-ICE

This is a reimplementation of the toolkit developed by Gerrit Holl for atmlab.

TODO: Extend documentation and credits.
"""

from .common import *

__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-

"""This package provides file handler classes. The file handler classes provide
specialized reading (sometimes as well writing) methods for several data
formats."""

from .common import *  # noqa
from .cloudsat import *  # noqa
from .hoaps import *  # noqa
from .meteosat import *
from .ocean_rain import *  # noqa
from .tovs import *  # noqa


# -*- coding: utf-8 -*-

"""This module provides functionality for reading and writing ARTS XML files.
"""

import gzip
import glob
import itertools
import os
from os.path import isfile, join, basename, splitext, dirname

from . import read
from . import write

__all__ = [
    'load',
    'save',
    'load_directory',
    'load_indexed',
    'make_binary',
    'make_directory_binary',
]


def save(var, filename, precision='.7e', format='ascii', comment=None,
         parents=False):
    """Save a variable to an ARTS XML file.

    Args:
        var: Variable to be stored.
        filename (str): Name of output XML file.
            If the name ends in .gz, the file is compressed on the fly.
        precision (str): Format for output precision.
        format (str): Output format: 'ascii' (default) or 'binary'.
        comment (str): Comment string included in a tag above data.
        parents (bool): Create missing parent directories.

    Note:
        Python's gzip module is extremely slow in writing. Consider
        compressing files manually after writing them normally.

    Example:
        >>> x = numpy.array([1.,2.,3.])
        >>> typhon.arts.xml.save(x, 'myvector.xml')

    """
    if parents:
        os.makedirs(dirname(filename), exist_ok=True)

    if filename.endswith('.gz'):
        if format != 'ascii':
            raise RuntimeError(
                'For zipped files, the output format must be "ascii"')
        xmlopen = gzip.open
    else:
        xmlopen = open
    with xmlopen(filename, mode='wt', encoding='UTF-8') as fp:
        if format == 'binary':
            with open(filename + '.bin', mode='wb') as binaryfp:
                axw = write.ARTSXMLWriter(fp, precision=precision,
                                          binaryfp=binaryfp)
                axw.write_header()
                if comment is not None:
                    axw.write_comment(comment)
                axw.write_xml(var)
                axw.write_footer()
        elif format == 'ascii':
            axw = write.ARTSXMLWriter(fp, precision=precision)
            axw.write_header()
            if comment is not None:
                axw.write_comment(comment)
            axw.write_xml(var)
            axw.write_footer()
        else:
            raise RuntimeError('Unknown output format "{}".'.format(format))


def load(filename):
    """Load a variable from an ARTS XML file.

    The input file can be either a plain or gzipped XML file

    Args:
        filename (str): Name of ARTS XML file.

    Returns:
        Data from the XML file. Type depends on data in file.

    Example:
        >>> typhon.arts.xml.load('tests/reference/matrix.xml')
        array([[ 0.,  1.],
               [ 2.,  3.]])

    """
    # If file is not found, try the gzipped version.
    if not isfile(filename):
        if not isfile(filename + '.gz'):
            raise FileNotFoundError("No such file: '{}'".format(filename))
        else:
            filename += '.gz'

    if filename.endswith('.gz'):
        xmlopen = gzip.open
    else:
        xmlopen = open

    binaryfilename = filename + '.bin'
    with xmlopen(filename, 'rb') as fp:
        if isfile(binaryfilename):
            with open(binaryfilename, 'rb',) as binaryfp:
                return read.parse(fp, binaryfp).getroot().value()
        else:
            return read.parse(fp).getroot().value()


def load_directory(directory, exclude=None):
    """Load all XML files in a given directory.

    Search given directory  for files with ``.xml`` or ``.xml.gz`` extension
    and try to load them using :func:`load`.

    Parameters:
        directory (str): Path to the directory.
        exclude (Container[str]): Filenames to exclude.

    Returns:
        dict: Filenames without extension are keys for the file content.

    Example:
        Load all files in ``foo`` except for the lookup table in
        ``abs_lookup.xml.``

        >>> load_directory('foo', exclude=['abs_lookup.xml'])
    """
    def includefile(f):
        """Check if to include file."""
        return basename(f) not in exclude if exclude is not None else True

    def stripext(f):
        """Strip the extension of a filename."""
        return splitext(f)[0]

    # Create a generator yielding all XML files to load (not excluded).
    xmlfiles = filter(includefile, glob.iglob(join(directory, '*.xml')))

    # Remove extension from zipped files to keep dictionary keys clean.
    # The `load` function looks for zipped files anyway.
    gzfiles = filter(includefile, glob.iglob(join(directory, '*.xml.gz')))
    gzfiles = map(stripext, gzfiles)

    # Store XML file contents in a dictionary, using the filename as key.
    return {stripext(basename(f)): load(f)
            for f in itertools.chain(xmlfiles, gzfiles)}


def load_indexed(filename):
    """Load all indexed XML files matching the given filename.

    The function searches all files matching the pattern
    ``<filename>.<file_index>.xml`` or ``<filename>.<file_index>.xml.gz``.

    A list with the loaded file contents is returned. The list indices are
    equivalent to the file indices.

    Parameters:
        filename (str): Filename.

    Returns:
        list: List of file contents.

    Example:
        Load all files matching the pattern ``foo.<file_index>.xml``.

        >>> load_indexed_xml('foo')

    """
    iidx = -2  # Relative position of fileindex in splitted filename.

    # Get all files matching the indexed filename format.
    files = glob.glob('{}.*.xml'.format(filename))

    # If no files are found, try the gzipped version.
    if len(files) == 0:
        files = glob.glob('{}.*.xml.gz'.format(filename))
        iidx = -3  # Correct fileindex position for gzipped files.

    # Extract indices from filenames.
    maxindex = max(int(x.split('.')[iidx]) for x in files)

    # Pre-allocate a list according to the maximum index found.
    ret = (maxindex + 1) * [None]

    # Fill list with file contents (file index matching list index).
    for f in files:
        findex = int(f.split('.')[iidx])
        ret[findex] = load(f)

    return ret


def make_binary(filename, out='', absolute_out=False, parents=True):
    """Loads xml-file at filename and saves it back in binary format

    Parameters:
        filename (str): Filename path.
        out (str): Path to save the binary.  Empty causes overwrite of file.
        absolute_out (bool): If true, then write file to out-path rather than
            to the relative path out.  Does nothing if file is in the working
            folder and out is relative.
        parents (bool): Create missing parent directories.

    Returns:
        str: Path to the created binary file.

    Example:
        Load t_field.xml and save it back as binary it as ./binary/t_field.xml
        and ./binary/t_field.bin

        >>> make_binary('t_field.xml', out='binary')
        'binary/t_field.xml'
    """

    xml_data = load(filename)
    if absolute_out:
        outfile = join(out, basename(filename))
    else:
        outfile = join(dirname(filename), out, basename(filename))

    save(xml_data, outfile, format='binary', parents=parents)

    return outfile


def make_directory_binary(directory, out='', absolute_out=False, parents=True):
    """Loads xml-files in directory and saves them back in binary format

    Parameters:
        directory (str): Directory path.
        out (str): Path to save the binary.
        absolute_out (bool): If true, then write file to out-path rather than
            to the relative path out.  Does nothing if file is in the working
            folder and out is relative.
        parents (bool): Create missing parent directories.

    Returns:
        list[str]: Paths to the created binary files.

    Example:
        Load arts-xml-data/spectroscopy/cia/hitran2011/ and save it back as
        binary it at arts-xml-data-binary/spectroscopy/cia/hitran2011/

        >>> make_directory_binary('arts-xml-data/spectroscopy/cia/hitran2011',
            out='arts-xml-data-binary/spectroscopy/cia/hitran2011',
            absolute_out=True)
        ['arts-xml-data-binary/spectroscopy/cia/hitran2011/hitran_cia2012_adapted.xml']
    """

    directory_of_xmls = load_directory(directory)
    outfiles = []  # Empty list to store output filepaths.

    if absolute_out:
        get_outfile = join(out, '{entry}.xml')
    else:
        get_outfile = join(directory, out, '{entry}.xml')

    for entry in directory_of_xmls:
        outfile = get_outfile.format(entry=entry)
        save(directory_of_xmls[entry],
             outfile,
             format='binary',
             parents=parents)
        outfiles.append(outfile)

    return outfiles
